@@ -23,50 +23,71 @@ extern int setcurdrv __P_((int, int));
 #ifdef	USEMKDEVH
 #include <sys/mkdev.h>
 #else
-# ifndef	major
-# define	major(n)	(((unsigned)(n) >> 8) & 0xff)
-# endif
-# ifndef	minor
-# define	minor(n)	((unsigned)((n) & 0xff))
+# ifdef	USEMKNODH
+# include <sys/mknod.h>
+# else
+#  ifdef	SVR4
+#  include <sys/sysmacros.h>
+#   ifndef	major
+#   define	major(n)	(((unsigned long)(n) >> 18) & 0x3fff)
+#   endif
+#   ifndef	minor
+#   define	minor(n)	((unsigned long)((n) & 0x3ffff))
+#   endif
+#  else
+#   ifndef	major
+#   define	major(n)	(((unsigned)(n) >> 8) & 0xff)
+#   endif
+#   ifndef	minor
+#   define	minor(n)	((unsigned)((n) & 0xff))
+#   endif
+#  endif
 # endif
 #endif
 
-#ifndef	_NOARCHIVE
-extern char *archivefile;
-#endif
 extern bindtable bindlist[];
 extern functable funclist[];
-extern char *findpattern;
 #ifndef	_NOWRITEFS
 extern int writefs;
 #endif
 extern char *curfilename;
 extern char *origpath;
-#ifndef	_NOTREE
-extern char *treepath;
+#ifndef	_NOARCHIVE
+extern char archivedir[];
 #endif
+extern int win_x;
+extern int win_y;
+#ifndef	_NOCUSTOMIZE
+extern int custno;
+#endif
+extern int fd_restricted;
+extern int internal_status;
 
 static VOID NEAR pathbar __P_((VOID_A));
+static VOID NEAR statusbar __P_((VOID_A));
 static VOID NEAR stackbar __P_((VOID_A));
+static VOID NEAR sizebar __P_((VOID_A));
 #if	!MSDOS
 static char *NEAR putowner __P_((char *, uid_t));
 static char *NEAR putgroup __P_((char *, gid_t));
 #endif
+static VOID NEAR infobar __P_((VOID_A));
 static VOID NEAR calclocate __P_((int));
+#ifndef	_NOSPLITWIN
+static int NEAR listupwin __P_((char *));
+#endif
+static int NEAR searchmove __P_((int, char *));
 #ifndef	_NOPRECEDE
 static VOID readstatus __P_((VOID_A));
 #endif
 static int NEAR browsedir __P_((char *, char *));
 
-int columns = 0;
+int curcolumns = 0;
 int defcolumns = 0;
 int minfilename = 0;
-int filepos = 0;
 int mark = 0;
 int fnameofs = 0;
 int displaymode = 0;
-int dispmode = 0;
-int sorton = 0;
 int sorttype = 0;
 int chgorder = 0;
 int stackdepth = 0;
@@ -78,9 +99,6 @@ namelist filestack[MAXSTACK];
 char fullpath[MAXPATHLEN] = "";
 char *macrolist[MAXMACROTABLE];
 int maxmacro = 0;
-namelist *filelist = NULL;
-int maxfile = 0;
-int maxent = 0;
 int isearch = 0;
 char *helpindex[10] = {
 #ifdef	_NOTREE
@@ -103,6 +121,11 @@ u_long flaglist[] = {
 	UF_APPEND, UF_IMMUTABLE, UF_NOUNLINK
 };
 #endif
+winvartable winvar[MAXWINDOWS];
+#ifndef	_NOSPLITWIN
+int windows = WINDOWS;
+int win = 0;
+#endif
 
 static u_short modelist[] = {
 	S_IFDIR, S_IFLNK, S_IFSOCK, S_IFIFO
@@ -117,7 +140,12 @@ static u_char colorlist[] = {
 #ifndef	_NOPRECEDE
 char *precedepath = NULL;
 static int maxstat = 0;
+static int haste = 0;
 #endif
+static int search_x = -1;
+static int search_y = -1;
+int calc_x = -1;
+int calc_y = -1;
 
 
 static VOID NEAR pathbar(VOID_A)
@@ -156,7 +184,7 @@ VOID helpbar(VOID_A)
 
 	for (i = 0; i < 10; i++) {
 		locate(ofs + (width + 1) * i + (i / 5) * 3, LHELP);
-		len = (width - strlen(helpindex[i])) / 2;
+		len = (width - (int)strlen(helpindex[i])) / 2;
 		putterm(t_standout);
 		for (j = 0; j < len; j++) putch2(' ');
 		kanjiputs2(helpindex[i], width - len, 0);
@@ -166,10 +194,9 @@ VOID helpbar(VOID_A)
 	tflush();
 }
 
-VOID statusbar(max)
-int max;
+static VOID NEAR statusbar(VOID_A)
 {
-	char *str[6];
+	char *str[6], buf[4 + 1];
 	int width;
 
 	locate(0, LSTATUS);
@@ -179,20 +206,29 @@ int max;
 	putterm(t_standout);
 	cputs2("Page:");
 	putterm(end_standout);
-	cprintf2("%2d/%2d ",
-		filepos / FILEPERPAGE + 1, (max - 1) / FILEPERPAGE + 1);
+	cputs2(ascnumeric(buf, filepos / FILEPERPAGE + 1, 2, 2));
+	putch2('/');
+	cputs2(ascnumeric(buf, (maxfile - 1) / FILEPERPAGE + 1, 2, 2));
+	putch2(' ');
 
 	locate(CMARK, LSTATUS);
 	putterm(t_standout);
 	cputs2("Mark:");
 	putterm(end_standout);
-	cprintf2("%4d/%4d ", mark, max);
+	cputs2(ascnumeric(buf, mark, 4, 4));
+	putch2('/');
+	cputs2(ascnumeric(buf, maxfile, 4, 4));
+	putch2(' ');
 
 	locate(CSORT, LSTATUS);
 	putterm(t_standout);
 	cputs2("Sort:");
 	putterm(end_standout);
 
+#ifndef	_NOPRECEDE
+	if (haste) kanjiputs(OMIT_K);
+	else
+#endif
 	if (sorton & 7) {
 		str[0] = ONAME_K;
 		str[1] = OEXT_K;
@@ -279,12 +315,12 @@ static VOID NEAR stackbar(VOID_A)
 	tflush();
 }
 
-VOID sizebar(VOID_A)
+static VOID NEAR sizebar(VOID_A)
 {
 	char buf[14 + 1];
 	long total, fre;
 
-	if (!sizeinfo) return;
+	if (!sizeinfo || !*fullpath) return;
 
 	locate(0, LSIZE);
 	putterm(l_clear);
@@ -293,8 +329,10 @@ VOID sizebar(VOID_A)
 	putterm(t_standout);
 	cputs2("Size:");
 	putterm(end_standout);
-	cprintf2("%14.14s/", inscomma(buf, marksize, 3, 14));
-	cprintf2("%14.14s ", inscomma(buf, totalsize, 3, 14));
+	cputs2(ascnumeric(buf, marksize, 3, 14));
+	putch2('/');
+	cputs2(ascnumeric(buf, totalsize, 3, 14));
+	putch2(' ');
 
 	getinfofs(".", &total, &fre);
 
@@ -302,15 +340,23 @@ VOID sizebar(VOID_A)
 	putterm(t_standout);
 	cputs2("Total:");
 	putterm(end_standout);
-	if (total < 0) cprintf2("%15.15s ", "?");
-	else cprintf2("%12.12s KB ", inscomma(buf, total, 3, 12));
+	if (total < 0) cputs2(ascnumeric(buf, total, 3, 15));
+	else {
+		cputs2(ascnumeric(buf, total, 3, 12));
+		cputs2(" KB");
+	}
+	putch2(' ');
 
 	locate(CFREE, LSIZE);
 	putterm(t_standout);
 	cputs2("Free:");
 	putterm(end_standout);
-	if (fre < 0) cprintf2("%15.15s ", "?");
-	else cprintf2("%12.12s KB ", inscomma(buf, fre, 3, 12));
+	if (fre < 0) cputs2(ascnumeric(buf, fre, 3, 15));
+	else {
+		cputs2(ascnumeric(buf, fre, 3, 12));
+		cputs2(" KB");
+	}
+	putch2(' ');
 
 	tflush();
 }
@@ -368,13 +414,13 @@ static char *NEAR putowner(buf, uid)
 char *buf;
 uid_t uid;
 {
-	char *s;
+	uidtable *up;
 	int i;
 
 	i = WOWNER;
 	if (uid == (uid_t)-1) while (--i >= 0) buf[i] = '?';
-	else if ((s = getpwuid2(uid))) strncpy3(buf, s, &i, 0);
-	else sprintf(buf, "%-*u", WOWNER, uid);
+	else if ((up = finduid(uid, NULL))) strncpy3(buf, up -> name, &i, 0);
+	else ascnumeric(buf, uid, -1, WOWNER);
 
 	return(buf);
 }
@@ -383,21 +429,19 @@ static char *NEAR putgroup(buf, gid)
 char *buf;
 gid_t gid;
 {
-	char *s;
+	gidtable *gp;
 	int i;
 
 	i = WGROUP;
 	if (gid == (gid_t)-1) while (--i >= 0) buf[i] = '?';
-	else if ((s = getgrgid2(gid))) strncpy3(buf, s, &i, 0);
-	else sprintf(buf, "%-*u", WGROUP, gid);
+	else if ((gp = findgid(gid, NULL))) strncpy3(buf, gp -> name, &i, 0);
+	else ascnumeric(buf, gid, -1, WGROUP);
 
 	return(buf);
 }
 #endif	/* !MSDOS */
 
-VOID infobar(list, no)
-namelist *list;
-int no;
+static VOID NEAR infobar(VOID_A)
 {
 	char *buf;
 	struct tm *tm;
@@ -406,16 +450,17 @@ int no;
 	int i;
 #endif
 
+	if (!filelist || maxfile <= 0) return;
 	locate(0, LINFO);
 
-	if (list[no].st_nlink < 0) {
+	if (filelist[filepos].st_nlink < 0) {
 		putterm(l_clear);
-		kanjiputs(list[no].name);
+		kanjiputs(filelist[filepos].name);
 		tflush();
 		return;
 	}
 #ifndef	_NOPRECEDE
-	if (!havestat(&(list[no]))) {
+	if (!havestat(&(filelist[filepos]))) {
 		putterm(l_clear);
 # if	MSDOS
 		len = WMODE + 4 + WDATE + 1 + WTIME + 1;
@@ -425,44 +470,44 @@ int no;
 # endif
 		width = n_lastcolumn - len;
 		locate(len, LINFO);
-		kanjiputs2(list[no].name, width, fnameofs);
+		kanjiputs2(filelist[filepos].name, width, fnameofs);
 		tflush();
 		return;
 	}
 #endif
 
 	buf = malloc2(n_lastcolumn * 2 + 1);
-	tm = localtime(&(list[no].st_mtim));
+	tm = localtime(&(filelist[filepos].st_mtim));
 
 #ifdef	HAVEFLAGS
 	if (isfileflg(dispmode)) {
-		putflags(buf, list[no].st_flags);
+		putflags(buf, filelist[filepos].st_flags);
 		buf[8] = ' ';
 		buf[9] = ' ';
 	}
 	else
 #endif
-	putmode(buf, (!isdisplnk(dispmode) && islink(&(list[no])))
-		? (S_IFLNK | 0777) : list[no].st_mode);
+	putmode(buf, (!isdisplnk(dispmode) && islink(&(filelist[filepos])))
+		? (S_IFLNK | 0777) : filelist[filepos].st_mode);
 	len = WMODE;
 
-	sprintf(buf + len, " %2d ", list[no].st_nlink);
+	sprintf(buf + len, " %2d ", filelist[filepos].st_nlink);
 	len += 4;
 
 #if	!MSDOS
-	putowner(buf + len, list[no].st_uid);
+	putowner(buf + len, filelist[filepos].st_uid);
 	len += WOWNER;
 	buf[len++] = ' ';
-	putgroup(buf + len, list[no].st_gid);
+	putgroup(buf + len, filelist[filepos].st_gid);
 	len += WGROUP;
 	buf[len++] = ' ';
 
-	if (isdev(&(list[no]))) sprintf(buf + len, "%3u, %3u ",
-		major(list[no].st_size) & 0xff,
-		minor(list[no].st_size) & 0xff);
+	if (isdev(&(filelist[filepos]))) sprintf(buf + len, "%3u, %3u ",
+		major(filelist[filepos].st_size) & 0xff,
+		minor(filelist[filepos].st_size) & 0xff);
 	else
 #endif
-	sprintf(buf + len, "%8ld ", (long)(list[no].st_size));
+	sprintf(buf + len, "%8ld ", (long)(filelist[filepos].st_size));
 	len = strlen(buf);
 
 	sprintf(buf + len, "%02d-%02d-%02d %02d:%02d ",
@@ -472,13 +517,13 @@ int no;
 	width = n_lastcolumn - len;
 
 #if	MSDOS
-	strncpy3(buf + len, list[no].name, &width, fnameofs);
+	strncpy3(buf + len, filelist[filepos].name, &width, fnameofs);
 #else	/* !MSDOS */
-	i = strncpy3(buf + len, list[no].name, &width, fnameofs);
-	if (islink(&(list[no]))) {
+	i = strncpy3(buf + len, filelist[filepos].name, &width, fnameofs);
+	if (islink(&(filelist[filepos]))) {
 		width += len;
 		len += i + 1;
-		if (strlen3(list[no].name) > fnameofs) {
+		if (strlen3(filelist[filepos].name) > fnameofs) {
 			if (len < width) buf[len++] = '-';
 			if (len < width) buf[len++] = '>';
 			len++;
@@ -487,7 +532,7 @@ int no;
 			char *tmp;
 
 			tmp = malloc2(width * 2 + 1);
-			i = Xreadlink(list[no].name, tmp, width * 2);
+			i = Xreadlink(filelist[filepos].name, tmp, width * 2);
 			if (i >= 0) {
 				tmp[i] = '\0';
 				strncpy3(buf + len, tmp, &width, 0);
@@ -508,26 +553,25 @@ VOID waitmes(VOID_A)
 	locate(0, LMESLINE);
 	putterm(l_clear);
 	kanjiputs(WAIT_K);
+	locate(win_x, win_y);
 	tflush();
 }
 
 static VOID NEAR calclocate(i)
 int i;
 {
-	int x, y;
-
 	i %= FILEPERPAGE;
-	x = (i / FILEPERLOW) * (n_column / FILEPERLINE);
-	y = i % FILEPERLOW;
+	calc_x = (i / FILEPERLOW) * (n_column / FILEPERLINE) + 1;
+	calc_y = i % FILEPERLOW + LFILETOP;
 #ifndef	_NOCOLOR
 	if (ansicolor == 2) {
 		chgcolor(ANSI_BLACK, 1);
-		locate(x, y + LFILETOP);
+		locate(calc_x - 1, calc_y);
 		putch2(' ');
 	}
 	else
 #endif
-	locate(x + 1, y + LFILETOP);
+	locate(calc_x, calc_y);
 }
 
 #if	MSDOS
@@ -542,16 +586,16 @@ int calcwidth(VOID_A)
 {
 	int width;
 
-	width = (n_column / columns) - 2 - 1;
-	if (columns < 2 && width - WIDTH1 >= minfilename) width -= WIDTH1;
-	if (columns < 3 && width - WIDTH2 >= minfilename) width -= WIDTH2;
-	if (columns < 4 && width - WIDTH3 >= minfilename) width -= WIDTH3;
+	width = (n_column / curcolumns) - 2 - 1;
+	if (curcolumns < 2 && width - WIDTH1 >= minfilename) width -= WIDTH1;
+	if (curcolumns < 3 && width - WIDTH2 >= minfilename) width -= WIDTH2;
+	if (curcolumns < 4 && width - WIDTH3 >= minfilename) width -= WIDTH3;
 	return(width);
 }
 
-VOID putname(list, no, standout)
+int putname(list, no, isstandout)
 namelist *list;
-int no, standout;
+int no, isstandout;
 {
 	char *buf;
 	struct tm *tm;
@@ -566,15 +610,15 @@ int no, standout;
 	calclocate(no);
 	putch2(ismark(&(list[no])) ? '*' : ' ');
 
-	if (standout < 0 && stable_standout) {
+	if (isstandout < 0 && stable_standout) {
 		putterm(end_standout);
 		calclocate(no);
-		return;
+		return((n_column / curcolumns) - 2 - 1);
 	}
 
 	width = calcwidth();
-	buf = malloc2((n_column / columns) * 2 + 1);
-	i = (standout && fnameofs > 0) ? fnameofs : 0;
+	buf = malloc2((n_column / curcolumns) * 2 + 1);
+	i = (isstandout && fnameofs > 0) ? fnameofs : 0;
 #ifdef	CODEEUC
 	wid = width;
 #endif
@@ -582,11 +626,15 @@ int no, standout;
 
 #ifndef	_NOPRECEDE
 	if (!havestat(&(list[no]))) {
-		if (standout > 0) putterm(t_standout);
+		if (isstandout > 0) putterm(t_standout);
 		kanjiputs(buf);
 		free(buf);
-		if (standout > 0) putterm(end_standout);
-		return;
+		if (isstandout > 0) putterm(end_standout);
+# ifdef	CODEEUC
+		return(wid);
+# else
+		return(width);
+# endif
 	}
 #endif
 
@@ -615,13 +663,13 @@ int no, standout;
 #endif
 	len = width;
 #ifdef	CODEEUC
-	width += (n_column / columns) - 2 - 1 - wid;
+	width += (n_column / curcolumns) - 2 - 1 - wid;
 #else
-	width = (n_column / columns) - 2 - 1;
+	width = (n_column / curcolumns) - 2 - 1;
 #endif
 
 	tm = NULL;
-	if (columns < 5 && len + WIDTH3 <= width) {
+	if (curcolumns < 5 && len + WIDTH3 <= width) {
 		if (isdir(&(list[no])))
 			sprintf(buf + len, " %*.*s", WSIZE, WSIZE, "<DIR>");
 #if	MSDOS || !defined (_NODOSDRIVE)
@@ -640,18 +688,18 @@ int no, standout;
 #endif
 		else sprintf(buf + len, " %*ld",
 			WSIZE, (long)(list[no].st_size));
-		if (strlen(buf + len) > WSIZE + 1)
+		if ((int)strlen(buf + len) > WSIZE + 1)
 			sprintf(buf + len, " %*.*s", WSIZE, WSIZE, "OVERFLOW");
 		len += WIDTH3;
 	}
-	if (columns < 3 && len + WIDTH2 <= width) {
+	if (curcolumns < 3 && len + WIDTH2 <= width) {
 		tm = localtime(&(list[no].st_mtim));
 		sprintf(buf + len, " %02d-%02d-%02d %2d:%02d",
 			tm -> tm_year % 100, tm -> tm_mon + 1, tm -> tm_mday,
 			tm -> tm_hour, tm -> tm_min);
 		len += WIDTH2;
 	}
-	if (columns < 2 && len + WIDTH1 <= width) {
+	if (curcolumns < 2 && len + WIDTH1 <= width) {
 		if (!tm) tm = localtime(&(list[no].st_mtim));
 		sprintf(buf + len, ":%02d", tm -> tm_sec);
 		len += 1 + WSECOND;
@@ -675,10 +723,10 @@ int no, standout;
 	}
 
 #ifndef	_NOCOLOR
-	if (ansicolor) chgcolor(color, standout > 0);
+	if (ansicolor) chgcolor(color, isstandout > 0);
 	else
 #endif
-	if (standout > 0) putterm(t_standout);
+	if (isstandout > 0) putterm(t_standout);
 	kanjiputs(buf);
 	free(buf);
 #ifndef	_NOCOLOR
@@ -689,7 +737,8 @@ int no, standout;
 	if (ansicolor) putterms(t_normal);
 	else
 #endif
-	if (standout > 0) putterm(end_standout);
+	if (isstandout > 0) putterm(end_standout);
+	return((n_column / curcolumns) - 2 - 1);
 }
 
 int listupfile(list, max, def)
@@ -697,7 +746,7 @@ namelist *list;
 int max;
 char *def;
 {
-	char *cp;
+	char *cp, buf[2 + 1];
 	int i, count, start, ret;
 
 	for (i = 0; i < FILEPERLOW; i++) {
@@ -706,14 +755,16 @@ char *def;
 	}
 
 	if (max <= 0) {
-		i = (n_column / columns) - 2 - 1;
+		i = (n_column / curcolumns) - 2 - 1;
 		locate(1, LFILETOP);
 		putch2(' ');
 		putterm(t_standout);
 		cp = NOFIL_K;
-		if (i > strlen(cp)) kanjiputs2(cp, i, 0);
-		else cprintf2("%-*.*s", i, i, "No Files");
+		if (i <= (int)strlen(cp)) cp = "NoFiles";
+		kanjiputs2(cp, i, 0);
 		putterm(end_standout);
+		win_x = i + 2;
+		win_y = LFILETOP;
 		return(0);
 	}
 
@@ -736,64 +787,161 @@ char *def;
 		ret = (max <= 1 || strcmp(list[1].name, "..")) ? 0 : 1;
 	}
 
-	locate(6, LSTATUS);
-	cprintf2("%2d", start / FILEPERPAGE + 1);
+	if (list == filelist) {
+		locate(6, LSTATUS);
+		cputs2(ascnumeric(buf, start / FILEPERPAGE + 1, 2, 2));
+	}
 
 	for (i = start, count = 0; i < max; i++, count++) {
 		if (count >= FILEPERPAGE) break;
-		putname(list, i, (i == ret) ? 1 : 0);
+		if (i != ret) putname(list, i, 0);
+		else {
+			win_x = putname(list, i, 1) + 1;
+			win_x += calc_x;
+			win_y = calc_y;
+		}
 	}
 
 	return(ret);
 }
 
-VOID movepos(list, max, old, fstat)
-namelist *list;
-int max, old;
+#ifndef	_NOSPLITWIN
+static int NEAR listupwin(def)
+char *def;
+{
+	int i, x, y, n, duplwin;
+
+	duplwin = win;
+	for (n = 1; n < windows; n++) {
+		locate(0, WHEADER - 1 + (n * (FILEPERLOW + 1)));
+		putterm(l_clear);
+		putch2(' ');
+		for (i = 2; i < n_column; i++) putch2('-');
+	}
+	for (n = WHEADER - 1 + (n * (FILEPERLOW + 1)); n < LSTACK; n++) {
+		locate(0, n);
+		putterm(l_clear);
+	}
+	n = -1;
+	for (win = 0; win < windows; win++) {
+		if (!filelist) continue;
+		if (win == duplwin) {
+			n = listupfile(filelist, maxfile, def);
+			x = win_x;
+			y = win_y;
+		}
+		else {
+			listupfile(filelist, maxfile, filelist[filepos].name);
+			putname(filelist, filepos, -1);
+		}
+	}
+	win = duplwin;
+
+	win_x = x;
+	win_y = y;
+	return(n);
+}
+
+int shutwin(n)
+int n;
+{
+	int i, duplwin;
+
+	duplwin = win;
+	win = n;
+#ifndef	_NOARCHIVE
+	while (archivefile) poparchdupl();
+	if (winvar[win].v_archivedir) {
+		free(winvar[win].v_archivedir);
+		winvar[win].v_archivedir = NULL;
+	}
+#endif
+	if (winvar[win].v_fullpath) {
+		free(winvar[win].v_fullpath);
+		winvar[win].v_fullpath = NULL;
+	}
+	if (filelist) {
+		for (i = 0; i < maxfile; i++)
+			if (filelist[i].name) free(filelist[i].name);
+		free(filelist);
+		filelist = NULL;
+	}
+	maxfile = maxent = filepos = sorton = dispmode = 0;
+	if (findpattern) {
+		free(findpattern);
+		findpattern = NULL;
+	}
+	win = duplwin;
+	return(n);
+}
+#endif	/* !_NOSPLITWIN */
+
+VOID movepos(old, fstat)
+int old;
 u_char fstat;
 {
-	if ((fstat & LISTUP) || old / FILEPERPAGE != filepos / FILEPERPAGE) {
-		filepos = listupfile(list, max, list[filepos].name);
+#ifndef	_NOSPLITWIN
+	if ((fstat & REWRITEMODE) >= REWIN) {
+		filepos = listupwin(filelist[filepos].name);
 		keyflush();
 	}
-	else if ((fstat & REWRITE) || old != filepos) {
-		if (old != filepos) putname(list, old, -1);
-		putname(list, filepos, 1);
+	else
+#endif
+	if (((fstat & REWRITEMODE) >= RELIST)
+	|| old / FILEPERPAGE != filepos / FILEPERPAGE) {
+		filepos =
+			listupfile(filelist, maxfile, filelist[filepos].name);
+		keyflush();
 	}
-	infobar(list, filepos);
+	else if (((fstat & REWRITEMODE) >= REWRITE) || old != filepos) {
+		if (old != filepos) putname(filelist, old, -1);
+		win_x = putname(filelist, filepos, 1) + 1;
+		win_x += calc_x;
+		win_y = calc_y;
+	}
+	infobar();
 }
 
 VOID rewritefile(all)
 int all;
 {
-#ifndef	_NOARCHIVE
-	if (archivefile) {
-		rewritearc(all);
-		return;
-	}
-#endif
+	if (!filelist || maxfile < 0) return;
 	if (all > 0) {
 		title();
 		helpbar();
-		infobar(filelist, filepos);
+		infobar();
 	}
 	sizebar();
-	statusbar(maxfile);
+	statusbar();
 	stackbar();
+#ifndef	_NOARCHIVE
+	if (archivefile) archbar(archivefile, archivedir);
+	else
+#endif
 	pathbar();
 	if (all >= 0) {
 #ifndef	_NOTREE
 		if (treepath) rewritetree();
 		else
 #endif
-		listupfile(filelist, maxfile, filelist[filepos].name);
-		locate(0, 0);
+#ifndef	_NOCUSTOMIZE
+		if (custno >= 0) rewritecust(all);
+		else
+#endif
+		if (filelist) {
+#ifdef	_NOSPLITWIN
+			listupfile(filelist, maxfile, filelist[filepos].name);
+#else
+			listupwin(filelist[filepos].name);
+#endif
+		}
 	}
+	locate(win_x, win_y);
+	tflush();
 }
 
-int searchmove(list, max, ch, buf)
-namelist *list;
-int max, ch;
+static int NEAR searchmove(ch, buf)
+int ch;
 char *buf;
 {
 	char *str[4];
@@ -823,12 +971,12 @@ char *buf;
 	}
 	else if (len && bindlist[i].f_func == SEARCH_BACK) {
 		if (n > 2) pos--;
-		else if (n > 1) pos = max - 1;
+		else if (n > 1) pos = maxfile - 1;
 		s = -1;
 	}
 	else {
 		if (n == 1) buf[len = 0] = '\0';
-		if (ch < ' ' || ch == C_DEL || ch >= K_MIN) isearch = 0;
+		if (isctl(ch) || ch >= K_MIN) isearch = 0;
 		else if (len < MAXNAMLEN - 1) {
 			buf[len++] = ch;
 			buf[len] = '\0';
@@ -837,8 +985,11 @@ char *buf;
 
 	if (!isearch) {
 		helpbar();
-		if (ch == CR) return(-1);
-		else if (ch != ESC) putterm(t_bell);
+		search_x = search_y = -1;
+		win_x = calc_x;
+		win_y = calc_y;
+		if (ch == K_CR) return(-1);
+		else if (ch != K_ESC) putterm(t_bell);
 		return(0);
 	}
 
@@ -849,19 +1000,22 @@ char *buf;
 	str[3] = SEABF_K;
 
 	i = 0;
-	if (pos >= 0 && pos < max) for (;;) {
-		if (!strnpathcmp2(list[pos].name, buf, len)) {
+	if (pos >= 0 && pos < maxfile) for (;;) {
+		if (!strnpathcmp2(filelist[pos].name, buf, len)) {
 			i = 1;
 			break;
 		}
 		pos += s;
-		if (pos < 0 || pos >= max) break;
+		if (pos < 0 || pos >= maxfile) break;
 	}
 
 	putterm(t_standout);
+	search_x = len;
 	len = kanjiputs(str[2 - s - i]);
 	putterm(end_standout);
 	kanjiputs2(buf, n_column - len - 1, 0);
+	if ((search_x += len) >= n_column) search_x = n_column - 1;
+	search_y = LHELP;
 	if (i) filepos = pos;
 	else if (n != 2 && ch != K_BS) putterm(t_bell);
 	isearch = s * (i + 2);
@@ -888,11 +1042,15 @@ static VOID readstatus(VOID_A)
 	if (keywaitfunc != readstatus) return;
 	if (isfile(&(filelist[i])) && filelist[i].st_size) sizebar();
 	if (i == filepos) {
-		putname(filelist, i, 1);
-		infobar(filelist, i);
+		win_x = putname(filelist, i, 1) + 1;
+		win_x += calc_x;
+		win_y = calc_y;
+		infobar();
 	}
 	else if (i / FILEPERPAGE == filepos / FILEPERPAGE)
 		putname(filelist, i, 0);
+	locate(win_x, win_y);
+	tflush();
 }
 #endif
 
@@ -906,28 +1064,20 @@ char *file, *def;
 	char *cp, buf[MAXNAMLEN + 1];
 	int ch, i, no, old;
 #ifndef	_NOPRECEDE
-	int haste;
-#endif
+	int dupsorton;
 
+	haste = 0;
+#endif
 #ifndef	_NOCOLOR
 	if (ansicolor) putterms(t_normal);
 #endif
 
-	maxfile = mark = 0;
+	mark = 0;
 	totalsize = marksize = 0;
-	buf[0] = '\0';
-	blocksize = (off_t)getblocksize(".");
 	chgorder = 0;
-	if (sorttype < 100) sorton = sorttype;
-#ifndef	_NOPRECEDE
-	haste = (!sorton && includepath(NULL, precedepath)) ? 1 : 0;
-#endif
 
-	if (!(dirp = Xopendir("."))) {
-		lostcwd(NULL);
-		if (!(dirp = Xopendir("."))) error(".");
-	}
 	fnameofs = 0;
+	win_x = win_y = 0;
 	waitmes();
 
 	re = NULL;
@@ -940,30 +1090,56 @@ char *file, *def;
 		re = regexp_init(findpattern, -1);
 	}
 
-	while ((dp = searchdir(dirp, re, cp))) {
-		if (ishidedot(dispmode) && *(dp -> d_name) == '.'
-		&& !isdotdir(dp -> d_name)) continue;
-		strcpy(file, dp -> d_name);
-		for (i = 0; i < stackdepth; i++)
-			if (!strcmp(file, filestack[i].name)) break;
-		if (i < stackdepth) continue;
+#ifndef	_NOARCHIVE
+	if (archivefile) {
+		maxfile = (*archivedir) ? 1 : 0;
+		blocksize = 1;
+		if (sorttype < 100) sorton = 0;
+		copyarcf(re, cp);
+		def = (*file) ? file : NULL;
+	}
+	else
+#endif
+	{
+		maxfile = 0;
+		blocksize = (off_t)getblocksize(".");
+		if (sorttype < 100) sorton = sorttype;
+#ifndef	_NOPRECEDE
+		if (includepath(NULL, fullpath, precedepath)) {
+			haste = 1;
+			dupsorton = sorton;
+			sorton = 0;
+		}
+#endif
+		if (!(dirp = Xopendir("."))) {
+			lostcwd(NULL);
+			if (!(dirp = Xopendir("."))) error(".");
+		}
 
-		addlist();
-		filelist[maxfile].name = strdup2(file);
-		filelist[maxfile].ent = maxfile;
-		filelist[maxfile].tmpflags = 0;
+		while ((dp = searchdir(dirp, re, cp))) {
+			if (ishidedot(dispmode) && *(dp -> d_name) == '.'
+			&& !isdotdir(dp -> d_name)) continue;
+			strcpy(buf, dp -> d_name);
+			for (i = 0; i < stackdepth; i++)
+				if (!strcmp(buf, filestack[i].name)) break;
+			if (i < stackdepth) continue;
+			addlist();
+			filelist[maxfile].name = strdup2(buf);
+			filelist[maxfile].ent = maxfile;
+			filelist[maxfile].tmpflags = 0;
 #ifndef	_NOPRECEDE
-		if (!haste)
+			if (!haste)
 #endif
-		{
-			if (getstatus(&(filelist[maxfile])) < 0) {
-				free(filelist[maxfile].name);
-				continue;
+			{
+				if (getstatus(&(filelist[maxfile])) < 0) {
+					free(filelist[maxfile].name);
+					continue;
+				}
 			}
+			maxfile++;
 		}
-		maxfile++;
+		Xclosedir(dirp);
 	}
-	Xclosedir(dirp);
 	regexp_free(re);
 
 	if (!maxfile) {
@@ -981,15 +1157,23 @@ char *file, *def;
 	title();
 	helpbar();
 	rewritefile(-1);
-
+#ifdef	_NOSPLITWIN
 	old = filepos = listupfile(filelist, maxfile, def);
+#else
+	old = filepos = listupwin(def);
+#endif
 	fstat = 0;
 	no = 1;
 
 	keyflush();
+	buf[0] = '\0';
 	for (;;) {
-		if (no) movepos(filelist, maxfile, old, fstat);
-		locate(0, 0);
+		if (no) movepos(old, fstat);
+		if (search_x >= 0 && search_y >= 0) {
+			win_x = search_x;
+			win_y = search_y;
+		}
+		locate(win_x, win_y);
 		tflush();
 #ifndef	_NOPRECEDE
 		if (haste) keywaitfunc = readstatus;
@@ -1007,7 +1191,7 @@ char *file, *def;
 
 		old = filepos;
 		if (isearch) {
-			no = searchmove(filelist, maxfile, ch, buf);
+			no = searchmove(ch, buf);
 			if (no >= 0) {
 				fnameofs = 0;
 				continue;
@@ -1018,8 +1202,11 @@ char *file, *def;
 		no = (bindlist[i].d_func < 255 && isdir(&(filelist[filepos])))
 			? (int)(bindlist[i].d_func)
 			: (int)(bindlist[i].f_func);
-		fstat = (no <= NO_OPERATION) ? funclist[no].stat
-			: (KILLSTK | RELIST);
+		if (no < FUNCLISTSIZ) fstat = funclist[no].stat;
+#ifndef	_NOARCHIVE
+		else if (archivefile) continue;
+#endif
+		else fstat = (KILLSTK | RESCRN | REWIN);
 
 		if (fstat & KILLSTK) {
 			if (stackdepth > 0) {
@@ -1041,48 +1228,88 @@ char *file, *def;
 					filelist[filepos].name);
 				stackbar();
 			}
+#ifndef	_NOARCHIVE
+			else if (archivefile);
+#endif
 #ifndef	_NOWRITEFS
 			else if (chgorder && writefs < 1 && no != WRITE_DIR
+			&& !fd_restricted
 			&& (i = writablefs(".")) > 0 && underhome(NULL) > 0) {
 				chgorder = 0;
-				if (yesno(WRTOK_K))
-					arrangedir(filelist, maxfile, i);
+				if (yesno(WRTOK_K)) arrangedir(i);
 			}
 #endif
 		}
 
 		curfilename = filelist[filepos].name;
-		if (maxfile <= 0 && !(fstat & NO_FILE)) no = 0;
-		else if (no <= NO_OPERATION)
-			no = (*funclist[no].func)(filelist, &maxfile, NULL);
-		else {
-			no = execusercomm(macrolist[no - NO_OPERATION - 1],
-				filelist[filepos].name,
-				filelist, &maxfile, 0, 0);
-			if (no < -1) no = 0;
+		if ((maxfile <= 0 && !(fstat & NO_FILE))
+#ifndef	_NOARCHIVE
+		|| (archivefile && !(fstat & ARCH))
+#endif
+		) no = 0;
+		else if (no < FUNCLISTSIZ) {
+			if (!fd_restricted || !(fstat & RESTRICT))
+				no = (*funclist[no].func)(NULL);
+			else {
+				warning(0, RESTR_K);
+				no = 0;
+			}
 		}
+		else {
+			no = execusercomm(macrolist[no - FUNCLISTSIZ],
+				filelist[filepos].name, 0, 0, 0);
+			no = (no < 0) ? 1 :
+				(internal_status >= -1) ? internal_status: 4;
+		}
+#ifndef	_NOPRECEDE
+		if (sorton) haste = 0;
+#endif
 
-		if (no < 0) break;
+		if (no < 0 || no >= 4) break;
 		if (no == 1 || no == 3) helpbar();
 		if (no < 2) {
 			fstat = 0;
 			continue;
 		}
-		else if (no >= 4) {
-			no -= 4;
-			strcpy(file, filelist[filepos].name);
-			break;
-		}
 
-		if (!(fstat & REWRITE) || old != filepos) fnameofs = 0;
-		if ((fstat & RELIST) == RELIST) {
+		if ((fstat & REWRITEMODE) != REWRITE || old != filepos)
+			fnameofs = 0;
+		if (fstat & RESCRN) {
 			title();
 			helpbar();
 			rewritefile(-1);
 		}
 	}
 
-	for (i = 0; i < maxfile; i++) free(filelist[i].name);
+#ifndef	_NOARCHIVE
+	if (archivefile) {
+		if (no < 0 || (no > 4 && archchdir(file) < 0)) {
+			poparchdupl();
+			strcpy(file, filelist[filepos].name);
+		}
+		else if (no == 4) {
+			if (filepos < 0) *file = '\0';
+			else strcpy(file, filelist[filepos].name);
+		}
+		no = 0;
+	}
+	else
+#endif
+	if (no >= 4) {
+		no -= 4;
+		strcpy(file, filelist[filepos].name);
+	}
+#ifndef	_NOARCHIVE
+	if (!archivefile)
+#endif
+	for (i = 0; i < maxfile; i++) {
+		free(filelist[i].name);
+		filelist[i].name = NULL;
+	}
+	maxfile = filepos = 0;
+#ifndef	_NOPRECEDE
+	if (haste && !sorton) sorton = dupsorton;
+#endif
 	return(no);
 }
 
@@ -1093,7 +1320,42 @@ char *cur;
 	char *cp, *tmp, *def;
 	int i, ischgdir;
 
-	findpattern = def = NULL;
+	for (i = 0; i < MAXWINDOWS; i++) {
+#ifndef	_NOSPLITWIN
+		win = i;
+		winvar[i].v_fullpath = NULL;
+# ifndef	_NOARCHIVE
+		winvar[i].v_archivedir = NULL;
+# endif
+#endif
+#ifndef	_NOARCHIVE
+		archduplp = NULL;
+		archivefile = NULL;
+		archtmpdir = NULL;
+		launchp = NULL;
+		arcflist = NULL;
+		maxarcf = 0;
+# ifndef	_NODOSDRIVE
+		archdrive = 0;
+# endif
+#endif
+#ifndef	_NOTREE
+		treepath = NULL;
+#endif
+		findpattern = NULL;
+		filelist = NULL;
+		maxfile = maxent = filepos = sorton = dispmode = 0;
+	}
+#ifndef	_NOSPLITWIN
+	win = 0;
+#endif
+
+	strcpy(file, ".");
+	sorton = sorttype % 100;
+	dispmode = displaymode;
+	curcolumns = defcolumns;
+
+	def = NULL;
 	if (cur) {
 		cp = evalpath(strdup2(cur), 1);
 #if	MSDOS
@@ -1108,9 +1370,9 @@ char *cur;
 		tmp = strcatdelim(fullpath);
 #else	/* !MSDOS */
 		if (
-#ifndef	_NODOSDRIVE
+# ifndef	_NODOSDRIVE
 		_dospath(cp) ||
-#endif
+# endif
 		*cp == _SC_) *(tmp = fullpath) = '\0';
 		else tmp = strcatdelim(fullpath);
 #endif	/* !MSDOS */
@@ -1157,13 +1419,6 @@ char *cur;
 	_chdir2(fullpath);
 #endif
 
-	filelist = NULL;
-	maxent = 0;
-	strcpy(file, ".");
-	sorton = sorttype % 100;
-	dispmode = displaymode;
-	columns = defcolumns;
-
 	for (;;) {
 		if (!def && !strcmp(file, "..")) {
 			if ((cp = strrdelim(fullpath, 1))) cp++;
@@ -1186,7 +1441,13 @@ char *cur;
 			}
 		}
 
+#ifdef	_NOARCHIVE
 		ischgdir = browsedir(file, def);
+#else
+		do {
+			ischgdir = browsedir(file, def);
+		} while (archivefile);
+#endif
 		if (ischgdir < 0) break;
 		if (ischgdir) def = NULL;
 		else {
@@ -1196,6 +1457,12 @@ char *cur;
 		}
 	}
 
+#ifdef	_NOSPLITWIN
 	if (filelist) free(filelist);
 	if (findpattern) free(findpattern);
+#else
+	for (i = 0; i < MAXWINDOWS; i++) shutwin(i);
+#endif
+	locate(0, n_line - 1);
+	tflush();
 }
