@@ -19,11 +19,16 @@
 #include <stdlib.h>
 #endif
 
+#if	FD < 2
+#define	_NOORIGSHELL
 #define	_NOSPLITWIN
 #define	_NOPRECEDE
+#define	_NOCUSTOMIZE
+#define	_NOEXTRACOPY
 #define	_NOUSEHASH
 #define	_NOORIGGLOB
 #define	_NOKANJIFCONV
+#endif	/* FD < 2 */
 
 #include "types.h"
 #include "pathname.h"
@@ -34,19 +39,32 @@ extern char *_mtrace_file;
 
 
 #if	MSDOS
+# if	FD >= 2
+#define	RUNCOMFILE	"~\\fd2.rc"
+# else
 #define	RUNCOMFILE	"~\\fd.rc"
+# endif
 #define	HISTORYFILE	"~\\fd.hst"
-#define	CMDLINE_DELIM	"\t ;<>|"
-#define	METACHAR	" \"'*,;<=>?[]|"
-#define	DQ_METACHAR	"\"$"
 #else
+# if	FD >= 2
+#define	RUNCOMFILE	"~/.fd2rc"
+# else
 #define	RUNCOMFILE	"~/.fdrc"
+# endif
 #define	HISTORYFILE	"~/.fd_history"
-#define	CMDLINE_DELIM	"\t !&:;()<>|"
-#define	METACHAR	" \"#$&'()*:;<=>?[\\]^`|"
-#define	DQ_METACHAR	"\"$\\`"
 #endif
 
+#if	MSDOS && defined (_NOORIGSHELL)
+/*	Using COMMAND.COM */
+#define	SHELL_OPERAND	"|"
+#define	CMDLINE_DELIM	"\t\n <>|"
+#else
+#define	SHELL_OPERAND	"&;(`|"
+#define	CMDLINE_DELIM	"\t\n &;()<>|"
+#endif
+
+#define	FDSHELL		"fdsh"
+
 /****************************************************************
  *	If you don't like the following tools as each uses,	*
  *	you should rewrite another suitable command name.	*
@@ -63,7 +81,10 @@ extern char *_mtrace_file;
 #define	DISPLAYMODE	0
 #define	SORTTREE	0
 #define	WRITEFS		0
+#define	IGNORECASE	0
+#define	INHERITCOPY	0
 #define	ADJTTY		0
+#define	USEGETCURSOR	0
 #define	WINDOWS		1
 #define	COLUMNS		2
 #define	MINFILENAME	12
@@ -93,8 +114,14 @@ extern char *_mtrace_file;
 #define	OJUNETPATH	""
 #define	HEXPATH		""
 #define	CAPPATH		""
+#define	UTF8PATH	""
 #define	NOCONVPATH	""
+#if	FD >= 2
+#define	PROMPT		"$ "
+#else
 #define	PROMPT		"sh#"
+#endif
+#define	PROMPT2		"> "
 
 
 /****************************************************************
@@ -108,10 +135,13 @@ extern char *_mtrace_file;
 #define	MAXFUNCTABLE	32
 #define	MAXFUNCLINES	16
 #define	MAXLINESTR	255
+#if	MSDOS
+#define	MAXCOMMSTR	(128 - 2)
+#else
 #define	MAXCOMMSTR	1023
+#endif
 #define	MAXSELECTSTRS	16
 #define	MAXSTACK	5
-#define	MAXARGS		128
 #define	MAXWINDOWS	2
 
 #ifdef	_NOSPLITWIN
