@@ -4,10 +4,11 @@
  *	Information Module
  */
 
-#include <ctype.h>
+#include <errno.h>
 #include "fd.h"
 #include "term.h"
 #include "funcno.h"
+#include "kctype.h"
 #include "kanji.h"
 
 #ifdef	NOERRNO
@@ -21,10 +22,7 @@ extern int errno;
 #if	MSDOS
 #undef	USEVFSH
 #undef	USEMNTENTH
-#include <dos.h>
-#include <errno.h>
 #include "unixemu.h"
-extern int supportLFN __P_((char *));
 #else
 #include <sys/param.h>
 #include <sys/file.h>
@@ -34,6 +32,7 @@ extern int supportLFN __P_((char *));
 #include <mntent.h>
 typedef struct mntent	mnt_t;
 #define	getmntent2(fp, mntp)	getmntent(fp)
+#define	hasmntopt2(mntp, opt)	strstr2((mntp) -> mnt_opts, opt)
 #endif	/* USEMNTENTH */
 
 #ifdef	USEMNTTABH
@@ -42,7 +41,7 @@ typedef struct mntent	mnt_t;
 typedef struct mnttab	mnt_t;
 #define	setmntent		fopen
 #define	getmntent2(fp, mntp)	(getmntent(fp, mntp) ? NULL : mntp)
-#define	hasmntopt(mntp, opt)	strstr2((mntp) -> mnt_mntopts, opt)
+#define	hasmntopt2(mntp, opt)	strstr2((mntp) -> mnt_mntopts, opt)
 #define	endmntent		fclose
 #define	mnt_dir		mnt_mountp
 #define	mnt_fsname	mnt_special
@@ -59,7 +58,7 @@ typedef struct mnttab	mnt_t;
 #include <sys/vmount.h>
 #endif	/* USEMNTCTL */
 
-#if !defined (USEMOUNTH) && !defined (USEFSDATA)\
+#if	!defined (USEMOUNTH) && !defined (USEFSDATA) \
 && (defined (USEGETFSSTAT) || defined (USEMNTINFOR) || defined (USEMNTINFO))
 #include <sys/mount.h>
 #endif
@@ -68,7 +67,7 @@ typedef struct mnttab	mnt_t;
 #include <sys/fs_types.h>
 #endif
 
-#if	defined (USEGETFSSTAT) || defined (USEMNTCTL)\
+#if	defined (USEGETFSSTAT) || defined (USEMNTCTL) \
 || defined (USEMNTINFOR) || defined (USEMNTINFO) || defined (USEGETMNT)
 typedef struct _mnt_t {
 	char *mnt_fsname;
@@ -76,10 +75,10 @@ typedef struct _mnt_t {
 	char *mnt_type;
 	char *mnt_opts;
 } mnt_t;
-static FILE *setmntent __P_((char *, char *));
-static mnt_t *getmntent2 __P_((FILE *, mnt_t *));
-#define	hasmntopt(mntp, opt)	strstr2((mntp) -> mnt_opts, opt)
-#if	defined(USEMNTINFO) || defined(USEGETMNT)
+static FILE *NEAR setmntent __P_((char *, char *));
+static mnt_t *NEAR getmntent2 __P_((FILE *, mnt_t *));
+#define	hasmntopt2(mntp, opt)	strstr2((mntp) -> mnt_opts, opt)
+#if	defined (USEMNTINFO) || defined (USEGETMNT)
 #define	endmntent(fp)
 #else
 #define	endmntent(fp)		{ if (fp) free(fp); }
@@ -93,7 +92,7 @@ static int mnt_size = 0;
 typedef struct fstab	mnt_t;
 #define	setmntent(file, mode)	(FILE *)(setfsent(), NULL)
 #define	getmntent2(fp, mntp)	getfsent()
-#define	hasmntopt(mntp, opt)	strstr2((mntp) -> fs_mntops, opt)
+#define	hasmntopt2(mntp, opt)	strstr2((mntp) -> fs_mntops, opt)
 #define	endmntent(fp)		endfsent()
 #define	mnt_dir		fs_file
 #define	mnt_fsname	fs_spec
@@ -107,7 +106,19 @@ typedef struct _mnt_t {
 	char *mnt_type;
 	char *mnt_opts;
 } mnt_t;
-#define	hasmntopt(mntp, opt)	strstr2((mntp) -> mnt_opts, opt)
+#define	hasmntopt2(mntp, opt)	strstr2((mntp) -> mnt_opts, opt)
+# ifdef	PC98
+# define	PT_FAT12	0x81	/* 0x80 | 0x01 */
+# define	PT_FAT16	0x91	/* 0x80 | 0x11 */
+# define	PT_FAT32	0xa0	/* 0x80 | 0x20 */
+# define	PT_FAT16X	0xa1	/* 0x80 | 0x21 */
+# else
+# define	PT_FAT12	0x01
+# define	PT_FAT16	0x04
+# define	PT_FAT16X	0x06
+# define	PT_FAT32	0x0b
+# define	PT_FAT32LBA	0x0c
+# endif
 #endif
 
 
@@ -170,25 +181,41 @@ typedef struct _statfs_t {
 	long	f_bavail;
 	long	f_files;
 } statfs_t;
-static int statfs2 __P_((char *, statfs_t *));
+extern int unixstatfs __P_((char *, statfs_t *));
+#define	statfs2		unixstatfs
 #define	blocksize(fs)	(fs).f_bsize
 #endif
 
 extern VOID error __P_((char *));
 extern int _chdir2 __P_((char *));
+extern char *strcpy2 __P_((char *, char *));
 extern char *getwd2 __P_((VOID_A));
 extern VOID warning __P_((int, char *));
 #if	MSDOS || !defined (_NODOSDRIVE)
 extern int dospath __P_((char *, char *));
 #endif
-#if	!MSDOS && !defined (_NODOSDRIVE)
+#if	MSDOS && !defined (_NOUSELFN)
+# ifdef	LSI_C
+# define	toupper2	toupper
+# else	/* !LSI_C */
+# define	toupper2(c)	(uppercase[(u_char)(c)])
+extern u_char uppercase[256];
+# endif	/* !LSI_C */
+extern int supportLFN __P_((char *));
+# ifndef	_NODOSDRIVE
+extern int checkdrive __P_((int));
+# endif
+#endif	/* MSDOS && !_NOUSELFN */
+#ifndef	_NODOSDRIVE
 extern int dosstatfs __P_((int, char *));
 #endif
+extern char *malloc2 __P_((ALLOC_T));
+extern char *realloc2 __P_((VOID_P, ALLOC_T));
 extern char *strstr2 __P_((char *, char *));
+extern char *ascnumeric __P_((char *, long, int, int));
 extern int kanjiputs __P_((char *));
 extern int kanjiputs2 __P_((char *, int, int));
 extern int Xaccess __P_((char *, int));
-extern char *inscomma __P_((char *, long, int));
 extern int kanjiprintf __P_((CONST char *, ...));
 
 extern bindtable bindlist[];
@@ -196,6 +223,9 @@ extern functable funclist[];
 extern char fullpath[];
 extern int sizeinfo;
 extern char *distributor;
+#ifndef	_NODOSDRIVE
+extern int needbavail;
+#endif
 
 #ifndef	MOUNTED
 #define	MOUNTED		"/etc/mtab"
@@ -234,33 +264,53 @@ extern char *distributor;
 #ifndef	MNTTYPE_DOS7
 #define	MNTTYPE_DOS7	"dos7"	/* MS-DOS on Win95 */
 #endif
-
-static int code2str __P_((char *, int));
-static int checkline __P_((int));
-static int getfsinfo __P_((char *, statfs_t *, mnt_t *));
-static int info1line __P_((int, char *, long, char *, char *));
-#if	!defined (USEFSDATA) || !defined (_NODOSDRIVE)
-static long calcKB __P_((long, long));
+#ifndef	MNTTYPE_FAT12
+#define	MNTTYPE_FAT12	"fat12"	/* MS-DOS */
 #endif
+#ifndef	MNTTYPE_FAT16
+#define	MNTTYPE_FAT16	"fat16(16bit sector)"	/* MS-DOS */
+#endif
+#ifndef	MNTTYPE_FAT16X
+#define	MNTTYPE_FAT16X	"fat16(32bit sector)"	/* MS-DOS */
+#endif
+#ifndef	MNTTYPE_FAT32
+#define	MNTTYPE_FAT32	"fat32"	/* Win98 */
+#endif
+
+static int NEAR code2str __P_((char *, int));
+static int NEAR checkline __P_((int));
+VOID help __P_((int));
+static int NEAR getfsinfo __P_((char *, statfs_t *, mnt_t *));
+int writablefs __P_((char *));
+long getblocksize __P_((char *));
+static int NEAR info1line __P_((int, char *, long, char *, char *));
+#if	!defined (USEFSDATA) || !defined (_NODOSDRIVE)
+static long NEAR calcKB __P_((long, long));
+#endif
+VOID getinfofs __P_((char *, long *, long *));
+int infofs __P_((char *));
 
 static int keycodelist[] = {
 	K_HOME, K_END, K_DL, K_IL, K_DC, K_IC,
 	K_BEG, K_EOL, K_NPAGE, K_PPAGE, K_CLR, K_ENTER,
 	K_BS, '\t', CR, ESC
 };
+#define	KEYCODESIZ	((int)(sizeof(keycodelist) / sizeof(int)))
 static char *keystrlist[] = {
 	"Home", "End", "DelLin", "InsLin", "Del", "Ins",
 	"Beg", "Eol", "PageDn", "PageUp", "Clr", "Enter",
 	"Bs", "Tab", "Ret", "Esc"
 };
 
-static int code2str(buf, code)
+static int NEAR code2str(buf, code)
 char *buf;
 int code;
 {
 	buf = buf + strlen(buf);
 	if (code >= K_F(1) && code <= K_F(20))
 		sprintf(buf, "F%-6d", code - K_F0);
+	else if ((code & ~0x7f) == 0x80 && isalpha(code & 0x7f))
+		sprintf(buf, "Alt-%c  ", code & 0x7f);
 	else if (code == K_UP) sprintf(buf, "%-7.7s", UPAR_K);
 	else if (code == K_DOWN) sprintf(buf, "%-7.7s", DWNAR_K);
 	else if (code == K_RIGHT) sprintf(buf, "%-7.7s", RIGAR_K);
@@ -268,24 +318,26 @@ int code;
 	else {
 		int i;
 
-		for (i = 0; keycodelist[i]; i++)
+		for (i = 0; i < KEYCODESIZ; i++)
 			if (code == keycodelist[i]) break;
-		if (keycodelist[i]) sprintf(buf, "%-7.7s", keystrlist[i]);
-		else if (code < ' ') sprintf(buf, "Ctrl-%c ", code + '@');
-		else if (code < 0x100) sprintf(buf, "'%c'    ", code);
+		if (i < KEYCODESIZ) sprintf(buf, "%-7.7s", keystrlist[i]);
+#ifndef	CODEEUC
+		else if (iskna(code)) sprintf(buf, "'%c'    ", code);
+#endif
+		else if (isctl(code))
+			sprintf(buf, "Ctrl-%c ", (code + '@') & 0x7f);
+		else if (!ismsb(code)) sprintf(buf, "'%c'    ", code);
 		else return(0);
 	}
 	return(1);
 }
 
-static int checkline(y)
+static int NEAR checkline(y)
 int y;
 {
-	if (y >= LSTACK) {
-		locate(0, y);
-		putterm(l_clear);
+	if (y >= LFILEBOTTOM) {
 		warning(0, HITKY_K);
-		y = WHEADER + 1;
+		y = LFILETOP + 1;
 	}
 	return(y);
 }
@@ -297,22 +349,22 @@ int mode;
 	int i, j, c, x, y;
 
 	if (distributor && *distributor) {
-		i = n_column - strlen(distributor) - 26;
+		i = n_column - strlen(distributor) - 24;
 		locate(i, LHELP);
 		putch2('[');
 		putterm(t_standout);
-		cprintf2(" Distributed by: <%s> ", distributor);
+		cprintf2(" Distributed by: %s ", distributor);
 		putterm(end_standout);
 		putch2(']');
 	}
 
 	x = 0;
-	y = WHEADER;
+	y = LFILETOP;
 	locate(0, y++);
 	putterm(l_clear);
 
-	for (i = 0; i < NO_OPERATION; i++) {
-		if (!funclist[i].hmes) continue;
+	for (i = 0; i < FUNCLISTSIZ ; i++) {
+		if (i == NO_OPERATION || i == WARNING_BELL) continue;
 		locate(x * (n_column / 2), y);
 		if (x ^= 1) putterm(l_clear);
 		else y++;
@@ -341,9 +393,9 @@ int mode;
 
 		y = checkline(y);
 	}
-	if (y > WHEADER + 1) {
+	if (y > LFILETOP + 1) {
 		if (x) y++;
-		for (; y <= LSTACK; y++) {
+		for (; y < LFILEBOTTOM; y++) {
 			locate(0, y);
 			putterm(l_clear);
 		}
@@ -353,19 +405,19 @@ int mode;
 
 #ifdef	USEMNTCTL
 /*ARGSUSED*/
-static FILE *setmntent(file, mode)
+static FILE *NEAR setmntent(file, mode)
 char *file, *mode;
 {
 	char *buf;
 
-	mntctl(MCTL_QUERY, sizeof(int), (struct vmount *)(&mnt_size));
-	buf = (char *)malloc2(mnt_size);
+	mntctl(MCTL_QUERY, sizeof(int), (struct vmount *)&mnt_size);
+	buf = malloc2(mnt_size);
 	mntctl(MCTL_QUERY, mnt_size, (struct vmount *)buf);
 	mnt_ptr = 0;
 	return((FILE *)buf);
 }
 
-static mnt_t *getmntent2(fp, mntp)
+static mnt_t *NEAR getmntent2(fp, mntp)
 FILE *fp;
 mnt_t *mntp;
 {
@@ -380,31 +432,32 @@ mnt_t *mntp;
 
 	if (mnt_ptr >= mnt_size) return(NULL);
 	buf = (char *)fp;
-	vmntp = (struct vmount *)(&buf[mnt_ptr]);
+	vmntp = (struct vmount *)&(buf[mnt_ptr]);
 
-	cp = &buf[mnt_ptr + vmntp -> vmt_data[VMT_OBJECT].vmt_off];
+	cp = &(buf[mnt_ptr + vmntp -> vmt_data[VMT_OBJECT].vmt_off]);
 	len = strlen(cp) + 1;
-	if (!(vmntp -> vmt_flags & MNT_REMOTE))
-		*(fsname = (char *)realloc2(fsname, len)) = '\0';
+	if (!(vmntp -> vmt_flags & MNT_REMOTE)) {
+		fsname = realloc2(fsname, len);
+		strcpy(fsname, cp);
+	}
 	else {
-		host = &buf[mnt_ptr + vmntp -> vmt_data[VMT_HOSTNAME].vmt_off];
+		host = &(buf[mnt_ptr
+			+ vmntp -> vmt_data[VMT_HOSTNAME].vmt_off]);
 		len += strlen(host) + 1;
-		fsname = (char *)realloc2(fsname, len);
-		strcpy(fsname, host);
-		strcat(fsname, ":");
+		fsname = realloc2(fsname, len);
+		strcpy(strcpy2(strcpy2(fsname, host), ":"), cp);
 	}
-	strcat(fsname, cp);
 
-	cp = &buf[mnt_ptr + vmntp -> vmt_data[VMT_STUB].vmt_off];
+	cp = &(buf[mnt_ptr + vmntp -> vmt_data[VMT_STUB].vmt_off]);
 	len = strlen(cp) + 1;
-	dir = (char *)realloc2(dir, len);
+	dir = realloc2(dir, len);
 	strcpy(dir, cp);
 
 	entp = getvfsbytype(vmntp -> vmt_gfstype);
 	if (entp) {
 		cp = entp -> vfsent_name;
 		len = strlen(cp) + 1;
-		type = (char *)realloc2(type, len);
+		type = realloc2(type, len);
 		strcpy(type, cp);
 	}
 	else if (type) {
@@ -424,7 +477,7 @@ mnt_t *mntp;
 
 #if	defined (USEMNTINFOR) || defined (USEMNTINFO) || defined (USEGETFSSTAT)
 /*ARGSUSED*/
-static FILE *setmntent(file, mode)
+static FILE *NEAR setmntent(file, mode)
 char *file, *mode;
 {
 	struct statfs *buf;
@@ -452,7 +505,7 @@ char *file, *mode;
 #define	MNT_RDONLY	M_RDONLY
 #endif
 
-static mnt_t *getmntent2(fp, mntp)
+static mnt_t *NEAR getmntent2(fp, mntp)
 FILE *fp;
 mnt_t *mntp;
 {
@@ -463,16 +516,16 @@ mnt_t *mntp;
 #ifdef	USEMNTINFO
 # ifdef	USEVFCNAME
 	struct vfsconf *conf;
-# define	getvfsbynumber(n)	((conf = getvfsbytype(n)) ? \
-					conf -> vfc_name : NULL)
+# define	getvfsbynumber(n)	((conf = getvfsbytype(n)) \
+					? conf -> vfc_name : NULL)
 # else	/* !USEVFCNAME */
 #  ifdef	USEFFSTYPE
 #  define	getvfsbynumber(n)	(buf[mnt_ptr].f_fstypename)
 #  else
 #   ifdef	INITMOUNTNAMES
 	static char *mnt_names[] = INITMOUNTNAMES;
-#   define	getvfsbynumber(n)	(((n) <= MOUNT_MAXTYPE) ? \
-					mnt_names[n] : NULL)
+#   define	getvfsbynumber(n)	(((n) <= MOUNT_MAXTYPE) \
+					? mnt_names[n] : NULL)
 #   else
 #   define	getvfsbynumber(n)	(NULL)
 #   endif
@@ -480,8 +533,8 @@ mnt_t *mntp;
 # endif	/* !USEVFCNAME */
 #else	/* !USEMNTINFO */
 # ifdef	USEGETFSSTAT
-# define	getvfsbynumber(n)	(((n) <= MOUNT_MAXTYPE) ? \
-					mnt_names[n] : NULL)
+# define	getvfsbynumber(n)	(((n) <= MOUNT_MAXTYPE) \
+					? mnt_names[n] : NULL)
 # endif
 #endif	/* !USEMNTINFO */
 	char *cp;
@@ -491,17 +544,17 @@ mnt_t *mntp;
 	buf = (struct statfs *)fp;
 
 	len = strlen(buf[mnt_ptr].f_mntfromname) + 1;
-	fsname = (char *)realloc2(fsname, len);
+	fsname = realloc2(fsname, len);
 	strcpy(fsname, buf[mnt_ptr].f_mntfromname);
 
 	len = strlen(buf[mnt_ptr].f_mntonname) + 1;
-	dir = (char *)realloc2(dir, len);
+	dir = realloc2(dir, len);
 	strcpy(dir, buf[mnt_ptr].f_mntonname);
 
 	cp = (char *)getvfsbynumber(buf[mnt_ptr].f_type);
 	if (cp) {
 		len = strlen(cp) + 1;
-		type = (char *)realloc2(type, len);
+		type = realloc2(type, len);
 		strcpy(type, cp);
 	}
 	else if (type) {
@@ -521,7 +574,7 @@ mnt_t *mntp;
 
 #ifdef	USEGETMNT
 /*ARGSUSED*/
-static FILE *setmntent(file, mode)
+static FILE *NEAR setmntent(file, mode)
 char *file, *mode;
 {
 	mnt_ptr = 0;
@@ -529,7 +582,7 @@ char *file, *mode;
 }
 
 /*ARGSUSED*/
-static mnt_t *getmntent2(fp, mntp)
+static mnt_t *NEAR getmntent2(fp, mntp)
 FILE *fp;
 mnt_t *mntp;
 {
@@ -543,15 +596,15 @@ mnt_t *mntp;
 		NOSTAT_MANY, NULL) <= 0) return(NULL);
 
 	len = strlen(buf.fd_req.devname) + 1;
-	fsname = (char *)realloc2(fsname, len);
+	fsname = realloc2(fsname, len);
 	strcpy(fsname, buf.fd_req.devname);
 
 	len = strlen(buf.fd_req.path) + 1;
-	dir = (char *)realloc2(dir, len);
+	dir = realloc2(dir, len);
 	strcpy(dir, buf.fd_req.path);
 
 	len = strlen(gt_names[buf.fd_req.fstype]) + 1;
-	type = (char *)realloc2(type, len);
+	type = realloc2(type, len);
 	strcpy(type, gt_names[buf.fd_req.fstype]);
 
 	mntp -> mnt_fsname = fsname;
@@ -563,56 +616,67 @@ mnt_t *mntp;
 }
 #endif	/* USEGETMNT */
 
-#if	MSDOS
-static int statfs2(path, buf)
-char *path;
-statfs_t *buf;
-{
-	union REGS regs;
-	int drv;
-
-	drv = dospath(path, NULL);
-	if (drv >= 'a' && drv <= 'z') drv -= 'a' - 1;
-	else if (drv >= 'A' && drv <= 'Z') drv -= 'A' - 1;
-	else drv = 0;
-
-	regs.x.ax = 0x3600;
-	regs.h.dl = (u_char)drv;
-	intdos(&regs, &regs);
-	if (regs.x.ax == 0xffff) {
-		errno = ENOENT;
-		return(-1);
-	}
-
-	buf -> f_bsize = (long)((u_short)(regs.x.ax) * (u_short)(regs.x.cx));
-	buf -> f_blocks = (u_short)(regs.x.dx);
-	buf -> f_bfree =
-	buf -> f_bavail = (u_short)(regs.x.bx);
-	buf -> f_files = -1;
-
-	return(0);
-}
-#endif
-
-static int getfsinfo(path, fsbuf, mntbuf)
+static int NEAR getfsinfo(path, fsbuf, mntbuf)
 char *path;
 statfs_t *fsbuf;
 mnt_t *mntbuf;
 {
 #if	MSDOS
-	char *dir;
+# ifndef	_NODOSDRIVE
+	int i;
+# endif
 
-	mntbuf -> mnt_fsname = "";
+	mntbuf -> mnt_fsname = "MSDOS";
 	mntbuf -> mnt_dir[0] = dospath(path, NULL);
-	strcpy(&(mntbuf -> mnt_dir[1]), ":\\");
-	mntbuf -> mnt_type = (supportLFN(mntbuf -> mnt_dir) > 0)
-		? MNTTYPE_DOS7 : MNTTYPE_PC;
+	mntbuf -> mnt_dir[1] = ':';
+	mntbuf -> mnt_dir[2] = _SC_;
+	mntbuf -> mnt_dir[3] = '\0';
+# ifdef	_NOUSELFN
+	mntbuf -> mnt_type = MNTTYPE_PC;
+# else	/* !_NOUSELFN */
+	switch (supportLFN(mntbuf -> mnt_dir)) {
+		case 2:
+			mntbuf -> mnt_type = MNTTYPE_FAT32;
+			break;
+		case 1:
+			mntbuf -> mnt_type = MNTTYPE_DOS7;
+			break;
+#  ifndef	_NODOSDRIVE
+		case -1:
+			mntbuf -> mnt_fsname = "LFN emurate";
+			mntbuf -> mnt_type = MNTTYPE_DOS7;
+			break;
+		case -2:
+			mntbuf -> mnt_fsname = "BIOS raw";
+			i = checkdrive(toupper2(mntbuf -> mnt_dir[0]) - 'A');
+			if (i == PT_FAT12) mntbuf -> mnt_type = MNTTYPE_FAT12;
+			else if (i == PT_FAT16)
+				mntbuf -> mnt_type = MNTTYPE_FAT16;
+			else if (i == PT_FAT16X)
+				mntbuf -> mnt_type = MNTTYPE_FAT16X;
+			else mntbuf -> mnt_type = MNTTYPE_FAT32;
+			break;
+#  endif
+		case -4:
+			errno = ENOENT;
+			return(-1);
+/*NOTREACHED*/
+			break;
+		default:
+			mntbuf -> mnt_type = MNTTYPE_PC;
+			break;
+	}
+# endif	/* !_NOUSELFN */
 	mntbuf -> mnt_opts = "";
-	dir = path;
+
+	if (statfs2(mntbuf -> mnt_dir, fsbuf) < 0) {
+		if (errno == ENOENT) return(-1);
+		memset((char *)fsbuf, 0xff, sizeof(statfs_t));
+	}
 #else	/* !MSDOS */
 	mnt_t *mntp, mnt;
 	FILE *fp;
-	char *dir, fsname[MAXPATHLEN + 1];
+	char *dir, fsname[MAXPATHLEN];
 	int len, match;
 #ifndef	_NODOSDRIVE
 	int drv;
@@ -625,7 +689,7 @@ mnt_t *mntbuf;
 #ifndef	_NODOSDRIVE
 	else if ((drv = dospath(path, NULL))) {
 		static char dosmntdir[4];
-		char buf[sizeof(long) * 3];
+		char buf[sizeof(long) * 3 + 1];
 
 		mntbuf -> mnt_fsname = "";
 		mntbuf -> mnt_dir = dosmntdir;
@@ -635,25 +699,27 @@ mnt_t *mntbuf;
 			? MNTTYPE_DOS7 : MNTTYPE_PC;
 		mntbuf -> mnt_opts = "";
 		if (dosstatfs(drv, buf) < 0) return(-1);
-		fsbuf -> f_bsize = *((long *)(&buf[sizeof(long) * 0]));
+		if (buf[sizeof(long) * 3] & 001)
+			mntbuf -> mnt_type = MNTTYPE_FAT32;
+		fsbuf -> f_bsize = *((long *)&(buf[sizeof(long) * 0]));
 #ifdef	USEFSDATA
 		fsbuf -> fd_req.btot = calcKB(fsbuf -> f_bsize,
-			*((long *)(&buf[sizeof(long) * 1])));
+			*((long *)&(buf[sizeof(long) * 1])));
 		fsbuf -> fd_req.bfree =
 		fsbuf -> fd_req.bfreen = calcKB(fsbuf -> f_bsize,
-			*((long *)(&buf[sizeof(long) * 2])));
-#else
-#ifdef	USESTATVFS
+			*((long *)&(buf[sizeof(long) * 2])));
+#else	/* !USEFSDATA */
+#ifdef	USESTATVFSH
 		fsbuf -> f_frsize = 0;
 #endif
-		fsbuf -> f_blocks = *((long *)(&buf[sizeof(long) * 1]));
+		fsbuf -> f_blocks = *((long *)&(buf[sizeof(long) * 1]));
 		fsbuf -> f_bfree =
-		fsbuf -> f_bavail = *((long *)(&buf[sizeof(long) * 2]));
-#endif
+		fsbuf -> f_bavail = *((long *)&(buf[sizeof(long) * 2]));
+#endif	/* !USEFSDATA */
 		fsbuf -> f_files = -1;
 		return(0);
 	}
-#endif
+#endif	/* !_NODOSDRIVE */
 	else if (_chdir2(path) < 0) return(-1);
 
 	if (!dir) {
@@ -662,7 +728,19 @@ mnt_t *mntbuf;
 		match = 0;
 
 		fp = setmntent(MOUNTED, "r");
-		while ((mntp = getmntent2(fp, &mnt))) {
+		for (;;) {
+#ifdef	DEBUG
+			_mtrace_file = "getmntent(start)";
+			mntp = getmntent2(fp, &mnt);
+			if (_mtrace_file) _mtrace_file = NULL;
+			else {
+				_mtrace_file = "getmntent(end)";
+				malloc(0);	/* dummy malloc */
+			}
+			if (!mntp) break;
+#else
+			if (!(mntp = getmntent2(fp, &mnt))) break;
+#endif
 			if ((len = strlen(mntp -> mnt_dir)) < match
 			|| strncmp(mntp -> mnt_dir, dir, len)
 			|| (mntp -> mnt_dir[len - 1] != _SC_
@@ -687,12 +765,12 @@ mnt_t *mntbuf;
 		errno = ENOENT;
 		return(-1);
 	}
-	memcpy(mntbuf, mntp, sizeof(mnt_t));
-#endif	/* !MSDOS */
+	memcpy((char *)mntbuf, (char *)mntp, sizeof(mnt_t));
 
 	if (statfs2(mntbuf -> mnt_dir, fsbuf) < 0
 	&& (path == dir || statfs2(path, fsbuf) < 0))
 		memset((char *)fsbuf, 0xff, sizeof(statfs_t));
+#endif	/* !MSDOS */
 	return(0);
 }
 
@@ -711,7 +789,7 @@ char *path;
 		return((drv >= 'A' && drv <= 'Z') ? 4 : 5);
 #endif
 	if (getfsinfo(path, &fsbuf, &mntbuf) < 0
-	|| hasmntopt(&mntbuf, "ro")) return(0);
+	|| hasmntopt2(&mntbuf, "ro")) return(0);
 
 #if	!MSDOS
 	if (!strcmp(mntbuf.mnt_type, MNTTYPE_43)
@@ -727,6 +805,7 @@ char *path;
 #endif
 	if (!strcmp(mntbuf.mnt_type, MNTTYPE_PC)) return(4);
 	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_DOS7)) return(5);
+	else if (!strcmp(mntbuf.mnt_type, MNTTYPE_FAT32)) return(5);
 	return(0);
 }
 
@@ -743,7 +822,7 @@ char *dir;
 	statfs_t fsbuf;
 	mnt_t mntbuf;
 #ifndef	DEV_BSIZE
-	struct stat buf;
+	struct stat st;
 #endif
 
 	if (!strcmp(dir, ".") && getfsinfo(dir, &fsbuf, &mntbuf) >= 0)
@@ -751,41 +830,19 @@ char *dir;
 #ifdef	DEV_BSIZE
 	return(DEV_BSIZE);
 #else
-	if (stat(dir, &buf) < 0) error(dir);
-	return((int)buf.st_size);
+	if (Xstat(dir, &st) < 0) error(dir);
+	return((int)(st.st_size));
 #endif
 #endif	/* !MSDOS */
 }
 
-char *inscomma(buf, n, digit)
-char *buf;
-long n;
-int digit;
-{
-	char tmp[12];
-	int i, j, len;
-
-	j = 0;
-	if (n < 0) buf[j++] = '?';
-	else {
-		sprintf(tmp, "%ld", n);
-		len = strlen(tmp);
-		for (i = j = 0; tmp[i]; i++) {
-			buf[j++] = tmp[i];
-			if (--len && !(len % digit)) buf[j++] = ',';
-		}
-	}
-	buf[j] = '\0';
-	return(buf);
-}
-
-static int info1line(y, ind, n, str, unit)
+static int NEAR info1line(y, ind, n, s, unit)
 int y;
 char *ind;
 long n;
-char *str, *unit;
+char *s, *unit;
 {
-	char buf[16];
+	char buf[MAXCOLSCOMMA(3) + 1];
 	int width;
 
 	locate(0, y);
@@ -793,19 +850,19 @@ char *str, *unit;
 	locate(n_column / 2 - 20, y);
 	kanjiprintf("%-20.20s", ind);
 	locate(n_column / 2 + 2, y);
-	if (str) {
+	if (s) {
 		width = n_column - (n_column / 2 + 2);
-		kanjiputs2(str, width, 0);
+		kanjiputs2(s, width, 0);
 	}
 	else {
-		cprintf2("%12.12s", inscomma(buf, n, 3));
+		cputs2(ascnumeric(buf, n, 3, MAXCOLSCOMMA(3)));
 		kanjiprintf(" %s", unit);
 	}
 	return(checkline(++y));
 }
 
 #if	!defined (USEFSDATA) || !defined (_NODOSDRIVE)
-static long calcKB(block, byte)
+static long NEAR calcKB(block, byte)
 long block, byte;
 {
 	if (block < 0 || byte <= 0) return(-1);
@@ -827,6 +884,9 @@ long *totalp, *freep;
 	statfs_t fsbuf;
 	mnt_t mntbuf;
 
+#ifndef	_NODOSDRIVE
+	needbavail = 1;
+#endif
 	if (getfsinfo(path, &fsbuf, &mntbuf) < 0)
 		*totalp = *freep = -1;
 	else {
@@ -838,6 +898,9 @@ long *totalp, *freep;
 		*freep = calcKB(fsbuf.f_bavail, blocksize(fsbuf));
 #endif
 	}
+#ifndef	_NODOSDRIVE
+	needbavail = 0;
+#endif
 }
 
 int infofs(path)
@@ -847,11 +910,17 @@ char *path;
 	mnt_t mntbuf;
 	int y;
 
+#ifndef	_NODOSDRIVE
+	needbavail = 1;
+#endif
 	if (getfsinfo(path, &fsbuf, &mntbuf) < 0) {
 		warning(ENOTDIR, path);
+#ifndef	_NODOSDRIVE
+		needbavail = 0;
+#endif
 		return(0);
 	}
-	y = WHEADER;
+	y = LFILETOP;
 	locate(0, y++);
 	putterm(l_clear);
 
@@ -874,12 +943,15 @@ char *path;
 #endif
 	y = info1line(y, FSBSZ_K, fsbuf.f_bsize, NULL, "bytes");
 	y = info1line(y, FSINO_K, fsbuf.f_files, NULL, UNIT_K);
-	if (y > WHEADER + 1) {
-		for (; y <= LSTACK; y++) {
+	if (y > LFILETOP + 1) {
+		for (; y < LFILEBOTTOM; y++) {
 			locate(0, y);
 			putterm(l_clear);
 		}
 		warning(0, HITKY_K);
 	}
+#ifndef	_NODOSDRIVE
+	needbavail = 0;
+#endif
 	return(1);
 }
