@@ -18,6 +18,9 @@
 #define	USEUTIME
 #define	USEMKTIME
 #define	SENSEPERSEC	20
+# if	defined (__TURBOC__) && defined (__WIN32__)
+# define	BCC32
+# endif
 # ifdef	__GNUC__
 # define	SIGFNCINT
 #  ifndef	DJGPP
@@ -26,7 +29,9 @@
 #  define	USEREGCOMP
 #  endif
 # else	/* !__GNUC__ */
-# define	NOUID_T
+#  ifndef	BCC32
+#  define	NOUID_T
+#  endif
 # define	NOFILEMODE
 # define	NOUNISTDH
 # endif	/* !__GNUC__ */
@@ -37,7 +42,9 @@ typedef unsigned short	u_short;
 typedef unsigned long	u_long;
 # endif
 # ifdef	__TURBOC__
-# define	NEAR		near
+#  ifndef	BCC32
+#  define	NEAR		near
+#  endif
 # define	FORCEDSTDC
 # define	SIGFNCINT
 typedef long	off_t;
@@ -46,6 +53,9 @@ typedef long	off_t;
 # define	NEAR		near
 # define	FORCEDSTDC
 # endif
+# if	defined (DJGPP) || defined (BCC32)
+# define	PROTECTED_MODE
+# endif
 #define	_SC_	'\\'
 #define	_SS_	"\\"
 #else	/* !MSDOS */
@@ -582,6 +592,32 @@ typedef long	off_t;
 #define	SIGFNCINT
 #endif
 
+#if	defined (__APPLE__) && defined (__MACH__) && !defined (OSTYPE)
+#define	BSD43
+#define	OSTYPE			"DARWIN"	/* aka Mac OS X */
+#define	USEMANLANG
+#define	BSDINSTALL
+#define	TARFROMPAX
+#define	TERMCAPLIB		"-lcurses"
+#define	DECLSIGLIST
+#define	DECLERRLIST
+#define	USELEAPCNT
+#define	USETERMIOS
+#define	USEMOUNTH
+#define	USEMNTINFO
+#define	USEFFSTYPE
+#define	USEREGCOMP
+#define	USESETENV
+#define	USEMKTIME
+#define	USESYSCONF
+#define	USEWAITPID
+#define	USESIGPMASK
+#define	USERESOURCEH
+#define	GETPGRPVOID
+#define	USESETPGID
+#define	SIGFNCINT
+#endif
+
 #if	defined (__386BSD__) && !defined (OSTYPE)
 #define	BSD43
 #define	OSTYPE			"ORG_386BSD"
@@ -638,7 +674,8 @@ typedef long	off_t;
 /*	BSDOS		;BSD/OS (BSDI) */
 /*	BOW		;BSD on Windows (ASCII) */
 /*	OPENBSD		;OpenBSD */
-/*	386BSD_ORG	;386BSD */
+/*	DARWIN		;Darwin (Apple) */
+/*	ORG_386BSD	;386BSD */
 
 /* #define CODEEUC	;kanji code type is EUC */
 /* #define USEMANLANG	;man(1) directory includes LANG environment value */
