@@ -6,7 +6,6 @@
 
 #ifdef	FD
 #include "fd.h"
-#define	EVALMACRO
 #else	/* !FD */
 #include <stdio.h>
 #include <string.h>
@@ -53,6 +52,7 @@
 # endif
 #else	/* !MSDOS */
 #include <pwd.h>
+#include <grp.h>
 #include <sys/file.h>
 #include <sys/param.h>
 #define	EXTWIDTH	0
@@ -89,7 +89,6 @@ extern char *strdup2 __P_((char *));
 extern char *strcpy2 __P_((char *, char *));
 extern char *strncpy2 __P_((char *, char *, int));
 extern int isdotdir __P_((char *));
-extern char *catargs __P_((char *[], int));
 # ifdef	_NOROCKRIDGE
 extern DIR *_Xopendir __P_((char *));
 # define	Xopendir	_Xopendir
@@ -98,6 +97,7 @@ extern DIR *Xopendir __P_((char *));
 # endif
 extern int Xclosedir __P_((DIR *));
 extern struct dirent *Xreaddir __P_((DIR *));
+extern char *Xgetwd __P_((char *));
 extern int Xstat __P_((char *, struct stat *));
 extern int Xaccess __P_((char *, int));
 # if	MSDOS || !defined (_NODOSDRIVE)
@@ -116,7 +116,6 @@ char *strdup2 __P_((char *));
 char *strcpy2 __P_((char *, char *));
 char *strncpy2 __P_((char *, char *, int));
 int isdotdir __P_((char *));
-static char *NEAR catargs __P_((char *[], int));
 # if	MSDOS
 DIR *Xopendir __P_((char *));
 int Xclosedir __P_((DIR *));
@@ -126,6 +125,15 @@ struct dirent *Xreaddir __P_((DIR *));
 # define	Xclosedir	closedir
 # define	Xreaddir	readdir
 # endif
+# ifdef DJGPP
+extern char *Xgetwd __P_((char *));
+# else
+#  ifdef	USEGETWD
+#  define	Xgetwd		(char *)getwd
+#  else
+#  define	Xgetwd(p)	(char *)getcwd(p, MAXPATHLEN)
+#  endif
+# endif
 #define	Xstat(f, s)	(stat(f, s) ? -1 : 0)
 #define	Xaccess		access
 #define	_dospath(s)	(isalpha(*(s)) && (s)[1] == ':')
@@ -152,16 +160,19 @@ static hashlist *NEAR newhash __P_((char *, char *, int, hashlist *));
 static hashlist *NEAR findhash __P_((char *, int));
 static VOID NEAR rmhash __P_((char *, int));
 #endif
+static int NEAR isexecute __P_((char *, int, int));
 #if	MSDOS
-static int NEAR extaccess __P_((char *, char *, int));
+static int NEAR extaccess __P_((char *, char *, int, int));
 #endif
-#if	!defined (FDSH) && !defined (_NOCOMPLETE)
+#ifndef	_NOCOMPLETE
 # if	!MSDOS
-static int NEAR completeuser __P_((char *, int, char ***));
+static int NEAR completeuser __P_((char *, int, int, char ***));
 # endif
-static int NEAR completefile __P_((char *, char *, int, int, int, char ***));
-static int NEAR completeexe __P_((char *, int, char ***));
+static int NEAR completefile __P_((char *, int, int, char ***,
+	char *, int, int));
+static int NEAR completeexe __P_((char *, int, int, char ***));
 #endif
+static int NEAR addmeta __P_((char *, char *, int, int));
 static char *NEAR evalshellparam __P_((int, int));
 static int NEAR replacevar __P_((char *, char **, int, int, int, int));
 static char *NEAR insertarg __P_((char *, int, char *, int, int));
@@ -200,10 +211,13 @@ static int NEAR evalhome __P_((char **, int, char **));
 #define	iskanji1(s, i)	(issjis1((s)[i]) && issjis2((s)[(i) + 1]))
 #endif
 
-#define	IFS_SET		" \t\n"
 #define	MAXLONGWIDTH	10
 #define	BUFUNIT		32
 #define	n_size(n)	((((n) / BUFUNIT) + 1) * BUFUNIT)
+#define	b_size(n, type) ((((n) / BUFUNIT) + 1) * BUFUNIT * sizeof(type))
+#define	b_realloc(ptr, n, type) \
+			(((n) % BUFUNIT) ? ((type *)(ptr)) \
+			: (type *)realloc2(ptr, b_size(n, type)))
 #define	c_malloc(size)	(malloc2((size) = BUFUNIT))
 #define	c_realloc(ptr, n, size) \
 			(((n) + 1 < (size)) \
@@ -223,8 +237,13 @@ char **(*getarglistfunc)__P_((VOID_A)) = NULL;
 char *(*getflagfunc)__P_((VOID_A)) = NULL;
 int (*checkundeffunc)__P_((char *, char *, int)) = NULL;
 VOID (*exitfunc)__P_((VOID_A)) = NULL;
+char *(*backquotefunc)__P_((char *)) = NULL;
 #if	!MSDOS
 int pathignorecase = 0;
+static uidtable *uidlist = NULL;
+static int maxuid = 0;
+static gidtable *gidlist = NULL;
+static int maxgid = 0;
 #endif	/* !MSDOS */
 #ifndef	LSI_C
 u_char uppercase[256] = {
@@ -405,26 +424,6 @@ char *name;
 	return(0);
 }
 
-static char *NEAR catargs(argv, delim)
-char *argv[];
-int delim;
-{
-	char *cp;
-	int i, len;
-
-	if (!argv) return(NULL);
-	for (i = len = 0; argv[i]; i++) len += strlen(argv[i]);
-	if (i < 1) return(strdup2(""));
-	len += (delim) ? i - 1 : 0;
-	cp = malloc2(len + 1);
-	len = strcpy2(cp, argv[0]) - cp;
-	for (i = 1; argv[i]; i++) {
-		if (delim) cp[len++] = delim;
-		len = strcpy2(cp + len, argv[i]) - cp;
-	}
-	return(cp);
-}
-
 # if	MSDOS
 DIR *Xopendir(dir)
 char *dir;
@@ -529,7 +528,6 @@ int d;
 }
 #endif	/* MSDOS || (FD && !_NODOSDRIVE) */
 
-#ifndef	FDSH
 char *strrdelim2(s, eol)
 char *s, *eol;
 {
@@ -569,7 +567,6 @@ int ptr;
 	return(1);
 #endif
 }
-#endif	/* !FDSH */
 
 char *strcatdelim(s)
 char *s;
@@ -582,6 +579,7 @@ char *s;
 	else
 #endif
 	i = 0;
+	if (!s[i]) return(&(s[i]));
 	if (s[i] == _SC_ && !s[i + 1]) return(&(s[i + 1]));
 
 	cp = NULL;
@@ -619,13 +617,17 @@ char *buf, *s1, *s2;
 	else
 #endif
 	i = 0;
-	if (s1[i] == _SC_ && !s1[i + 1]) *(cp = &(buf[i])) = _SC_;
+	if (!s1[i]) cp = &(buf[i]);
+	else if (s1[i] == _SC_ && !s1[i + 1]) {
+		cp = &(buf[i]);
+		*(cp++) = _SC_;
+	}
 	else {
 		cp = NULL;
 		for (; s1[i]; i++) {
 			buf[i] = s1[i];
 			if (s1[i] == _SC_) {
-				if (!cp) cp = &(buf[i]);
+				if (!cp) cp = &(buf[i]) + 1;
 				continue;
 			}
 			cp = NULL;
@@ -642,17 +644,42 @@ char *buf, *s1, *s2;
 				*cp = '\0';
 				return(cp);
 			}
-			*cp = _SC_;
+			*(cp++) = _SC_;
 		}
 	}
 	if (s2) {
 		len = MAXPATHLEN - 1 - (cp - buf);
-		for (i = 0; s2[i] && i < len; i++) *(++cp) = s2[i];
+		for (i = 0; s2[i] && i < len; i++) *(cp++) = s2[i];
 	}
-	*(++cp) = '\0';
+	*cp = '\0';
 	return(cp);
 }
 
+#if	!MSDOS
+int strcasecmp2(s1, s2)
+char *s1, *s2;
+{
+	int c1, c2;
+
+	for (;;) {
+		c1 = toupper2(*s1);
+		c2 = toupper2(*s2);
+		if (c1 != c2) return(c1 - c2);
+# ifndef	CODEEUC
+		if (issjis1(c1)) {
+			s1++;
+			s2++;
+			if (*s1 != *s2) return(*s1 - *s2);
+		}
+# endif
+		if (!*s1) break;
+		s1++;
+		s2++;
+	}
+	return(0);
+}
+#endif	/* !MSDOS */
+
 int strpathcmp2(s1, s2)
 char *s1, *s2;
 {
@@ -862,7 +889,7 @@ int len;
 # if	MSDOS
 	if (regcomp(re, s, REG_EXTENDED | REG_ICASE)) {
 # else
-	if (regcomp(re, s, REG_EXTENDED)) {
+	if (regcomp(re, s, REG_EXTENDED | (pathignorecase ? REG_ICASE : 0))) {
 # endif
 		free(re);
 		re = NULL;
@@ -883,7 +910,10 @@ int fname;
 VOID regexp_free(re)
 reg_t *re;
 {
-	if (re) regfree(re);
+	if (re) {
+		regfree(re);
+		free(re);
+	}
 }
 
 # else	/* !USEREGCOMP */
@@ -957,12 +987,13 @@ int len;
 
 			if (quote || meta) {
 #ifdef	BASHSTYLE
+	/* bash treats a meta character in "[]" as a character itself */
 				paren[plen++] = PMETA;
-				paren[plen++] =
-#if	!MSDOS
-					(!pathignorecase) ? s[i] :
-#endif
-					toupper2(s[i]);
+# if	!MSDOS
+				if (!pathignorecase) paren[plen++] = s[i];
+				else
+# endif
+				paren[plen++] = toupper2(s[i]);
 #endif
 			}
 			else if (s[i] == ']') {
@@ -971,22 +1002,49 @@ int len;
 					regexp_free(re);
 					return(NULL);
 				}
+#ifndef	BASHSTYLE
+	/* bash treats "[a-]]" as "[a-]" + "]" */
+				if (paren[plen - 1] == '-') {
+					paren[plen++] = s[i];
+					continue;
+				}
+#endif
 				paren[plen] = '\0';
 				cp = paren;
 				paren = NULL;
 			}
 			else if (!plen && s[i] == '!') paren[plen++] = '\0';
-#ifndef	BASHSTYLE
 			else if (s[i] == '-') {
-				if (!plen
-				|| !s[i + 1] || s[i + 1] == ']') {
+				if (i + 1 >= len) j = '\0';
+#if	!MSDOS
+				else if (!pathignorecase) j = s[i + 1];
+#endif
+				else j = toupper2(s[i + 1]);
+
+#ifdef	BASHSTYLE
+	/* bash does not allows "[b-a]" */
+				if (plen && j && j != ']'
+				&& j < paren[plen - 1]) {
+					free(paren);
+					regexp_free(re);
+					return(NULL);
+				}
+#else
+				if (plen && j && j < paren[plen - 1]) {
+					do {
+						if (i + 1 >= len) break;
+						i++;
+					} while (s[i + 1] != ']');
+				}
+	/* bash allows "[-a]" */
+				else if ((!plen && j != '-')) {
 					free(paren);
 					regexp_free(re);
 					return(NULL);
 				}
-				paren[plen++] = s[i];
-			}
 #endif
+				else paren[plen++] = s[i];
+			}
 			else {
 #ifdef	CODEEUC
 				if (isekana(s, i)) {
@@ -999,14 +1057,13 @@ int len;
 					paren[plen++] = s[i++];
 					paren[plen++] = s[i];
 				}
-				else paren[plen++] =
 #if	!MSDOS
-					(!pathignorecase) ? s[i] :
+				else if (!pathignorecase) paren[plen++] = s[i];
 #endif
-					toupper2(s[i]);
+				else paren[plen++] = toupper2(s[i]);
 			}
 		}
-		else if (!quote) switch(s[i]) {
+		else if (!quote) switch (s[i]) {
 			case '?':
 				cp = wildsymbol1;
 				break;
@@ -1034,11 +1091,10 @@ int len;
 				cp[j++] = s[i++];
 				cp[j++] = s[i];
 			}
-			else cp[j++] =
 #if	!MSDOS
-				(!pathignorecase) ? s[i] :
+			else if (!pathignorecase) cp[j++] = s[i];
 #endif
-				toupper2(s[i]);
+			else cp[j++] = toupper2(s[i]);
 			cp[j] = '\0';
 		}
 		re = (reg_t *)realloc2(re, (n + 2) * sizeof(reg_t));
@@ -1086,6 +1142,7 @@ char *s;
 		c2 = '\0';
 		for (; re[n1][i]; i++) {
 #ifdef	BASHSTYLE
+	/* bash treats a meta character in "[]" as a character itself */
 			if (re[n1][i] == PMETA && re[n1][i + 1]) i++;
 			if (re[n1][i] == '-' && re[n1][i + 1] && c2)
 #else
@@ -1339,6 +1396,13 @@ devino_t *ino;
 			dupino[n].ino = st.st_ino;
 #endif
 			if (!re) {
+				if (*(dp -> d_name) == '.') {
+					free(cp);
+#if	!MSDOS
+					free(dupino);
+#endif
+					continue;
+				}
 #if	MSDOS
 				argc = _evalwild(argc, argvp, s, cp, l);
 #else
@@ -1483,6 +1547,7 @@ hashlist **htable;
 		for (hp = htable[i]; hp; hp = hp -> next) {
 			*nextp = newhash(hp -> command,
 				strdup2(hp -> path), hp -> cost, NULL);
+			(*nextp) -> type = hp -> type;
 			(*nextp) -> hits = hp -> hits;
 			nextp = &((*nextp) -> next);
 		}
@@ -1522,13 +1587,27 @@ int n;
 }
 #endif	/* !_NOUSEHASH */
 
+static int NEAR isexecute(path, dirok, exe)
+char *path;
+int dirok, exe;
+{
+	struct stat st;
+	int d;
+
+	if (Xstat(path, &st) < 0) return(-1);
+	d = ((st.st_mode & S_IFMT) == S_IFDIR);
+	if (!exe) return(d);
+	if (d) return(dirok ? d : -1);
+	return(Xaccess(path, X_OK));
+}
+
 #if	MSDOS
-static int NEAR extaccess(path, ext, len)
+static int NEAR extaccess(path, ext, len, exe)
 char *path, *ext;
-int len;
+int len, exe;
 {
 	if (ext) {
-		if (Xaccess(path, F_OK) >= 0) {
+		if (isexecute(path, 0, exe) >= 0) {
 			if (!(strpathcmp(ext, "COM"))) return(0);
 			else if (!(strpathcmp(ext, "EXE"))) return(CM_EXE);
 			else if (!(strpathcmp(ext, "BAT"))) return(CM_BATCH);
@@ -1537,26 +1616,28 @@ int len;
 	else {
 		path[len++] = '.';
 		strcpy(path + len, "COM");
-		if (Xaccess(path, F_OK) >= 0) return(CM_ADDEXT);
+		if (isexecute(path, 0, exe) >= 0) return(CM_ADDEXT);
 		strcpy(path + len, "EXE");
-		if (Xaccess(path, F_OK) >= 0) return(CM_ADDEXT | CM_EXE);
+		if (isexecute(path, 0, exe) >= 0) return(CM_ADDEXT | CM_EXE);
 		strcpy(path + len, "BAT");
-		if (Xaccess(path, F_OK) >= 0) return(CM_ADDEXT | CM_BATCH);
+		if (isexecute(path, 0, exe) >= 0) return(CM_ADDEXT | CM_BATCH);
 	}
 	return(-1);
 }
 #endif
 
-int searchhash(hpp, com)
+int searchhash(hpp, com, search)
 hashlist **hpp;
-char *com;
+char *com, *search;
 {
-	char *cp, *next, *path;
-	int l, len, cost, size, ret;
+	char *cp, *tmp, *next, *path;
+	int len, dlen, cost, size, ret, recalc;
 #if	MSDOS
 	char *ext;
 #endif
 #ifndef	_NOUSEHASH
+	static char *searchcwd = NULL;
+	char buf[MAXPATHLEN];
 	int n;
 #endif
 
@@ -1568,11 +1649,11 @@ char *com;
 		len = strlen(com);
 		path = malloc2(len + EXTWIDTH + 1);
 		strcpy(path, com);
-		ret = extaccess(path, ext, len);
+		ret = extaccess(path, ext, len, 0);
 		free(path);
 		if (ret >= 0) return(ret | CM_FULLPATH);
 #else
-		if (Xaccess(com, F_OK) >= 0) return(CM_FULLPATH);
+		if (isexecute(com, 0, 0) >= 0) return(CM_FULLPATH);
 #endif
 		return(CM_NOTFOUND | CM_FULLPATH);
 	}
@@ -1582,7 +1663,9 @@ char *com;
 	n = calchash(com);
 	if ((*hpp = findhash(com, n))) {
 		path = (*hpp) -> path;
-		if (Xaccess(path, F_OK) >= 0) return((*hpp) -> type);
+		if (((*hpp) -> type & CM_RECALC)
+		&& (!searchcwd || strcmp(searchcwd, Xgetwd(buf))));
+		else if (isexecute(path, 0, 1) >= 0) return((*hpp) -> type);
 		rmhash(com, n);
 	}
 #endif
@@ -1593,43 +1676,55 @@ char *com;
 	path[0] = '.';
 	path[1] = _SC_;
 	strcpy(path + 2, com);
-	if ((ret = extaccess(path, ext, len + 2)) < 0) free(path);
+	if ((ret = extaccess(path, ext, len + 2, 1)) < 0) free(path);
 	else {
 # ifdef	_NOUSEHASH
 		*hpp = (hashlist *)path;
 # else
 		*hpp = newhash(com, path, 0, hashtable[n]);
 		hashtable[n] = *hpp;
-		(*hpp) -> type = (ret |= CM_HASH);
+		(*hpp) -> type = (ret |= (CM_HASH | CM_RECALC));
+		if (searchcwd) free(searchcwd);
+		searchcwd = strdup2(Xgetwd(buf));
 # endif
 		return(ret);
 	}
-#endif
-	if ((next = getvar("PATH", -1))) {
+#endif	/* MSDOS && !DISMISS_CURPATH */
+
+	recalc = 0;
+	if ((next = (search) ? search : getvar("PATH", -1))) {
 		len = strlen(com);
 		size = ret = 0;
 		path = NULL;
 		cost = 1;
 		for (cp = next; cp; cp = next) {
+			next = cp;
 #if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
-			if (_dospath(cp)) next = strchr(cp + 2, PATHDELIM);
-			else
+			if (_dospath(cp)) next += 2;
 #endif
-			next = strchr(cp, PATHDELIM);
-			l = (next) ? (next++) - cp : strlen(cp);
-			if (l + 1 + len + EXTWIDTH + 1 > size) {
-				size = l + len + 1 + EXTWIDTH + 1;
+			if (*next != _SC_) recalc = CM_RECALC;
+			next = strchr(next, PATHDELIM);
+			dlen = (next) ? (next++) - cp : strlen(cp);
+			if (!dlen) tmp = NULL;
+			else {
+				tmp = _evalpath(cp, cp + dlen, 1, 1);
+				dlen = strlen(tmp);
+			}
+			if (dlen + len + 1 + EXTWIDTH + 1 > size) {
+				size = dlen + len + 1 + EXTWIDTH + 1;
 				path = realloc2(path, size);
 			}
-			if (l) {
-				strncpy2(path, cp, l);
-				l = strcatdelim(path) - path;
+			if (tmp) {
+				strncpy2(path, tmp, dlen);
+				free(tmp);
+				dlen = strcatdelim(path) - path;
 			}
-			l = strcpy2(path + l, com) - path;
+			strncpy2(path + dlen, com, len);
+			dlen += len;
 #if	MSDOS
-			if ((ret = extaccess(path, ext, l)) >= 0) break;
+			if ((ret = extaccess(path, ext, dlen, 1)) >= 0) break;
 #else
-			if (Xaccess(path, F_OK) >= 0) break;
+			if (isexecute(path, 0, 1) >= 0) break;
 #endif
 			cost++;
 		}
@@ -1638,22 +1733,27 @@ char *com;
 			*hpp = (hashlist *)path;
 #else
 			*hpp = newhash(com, path, cost, hashtable[n]);
-			hashtable[n] = 	*hpp;
-			(*hpp) -> type = (ret |= CM_HASH);
+			hashtable[n] = *hpp;
+			(*hpp) -> type = (ret |= (CM_HASH | recalc));
+			if (searchcwd) free(searchcwd);
+			searchcwd = strdup2(Xgetwd(buf));
 #endif
 			return(ret);
 		}
+		if (path) free(path);
 	}
-	free(path);
+
 	return(CM_NOTFOUND);
 }
 
-char *searchpath(path)
-char *path;
+char *searchpath(path, search)
+char *path, *search;
 {
 	hashlist *hp;
+	int type;
 
-	if (searchhash(&hp, path) & CM_NOTFOUND) return(NULL);
+	if ((type = searchhash(&hp, path, search)) & CM_NOTFOUND) return(NULL);
+	if (type & CM_FULLPATH) return(strdup2(path));
 #ifdef	_NOUSEHASH
 	return((char *)hp);
 #else
@@ -1661,7 +1761,7 @@ char *path;
 #endif
 }
 
-#if	!defined (FDSH) && !defined (_NOCOMPLETE)
+#ifndef	_NOCOMPLETE
 char *finddupl(target, argc, argv)
 char *target;
 int argc;
@@ -1675,18 +1775,57 @@ char **argv;
 }
 
 #if	!MSDOS
-static int NEAR completeuser(name, argc, argvp)
+static int NEAR completeuser(name, len, argc, argvp)
 char *name;
-int argc;
+int len, argc;
 char ***argvp;
 {
 	struct passwd *pwd;
 	char *new;
-	int len, size;
+	int size;
 
 	len = strlen(name);
+# ifdef	DEBUG
+	_mtrace_file = "setpwent(start)";
 	setpwent();
-	while ((pwd =getpwent())) {
+	if (_mtrace_file) _mtrace_file = NULL;
+	else {
+		_mtrace_file = "setpwent(end)";
+		malloc(0);	/* dummy alloc */
+	}
+	for (;;) {
+		_mtrace_file = "getpwent(start)";
+		pwd = getpwent();
+		if (_mtrace_file) _mtrace_file = NULL;
+		else {
+			_mtrace_file = "getpwent(end)";
+			malloc(0);	/* dummy alloc */
+		}
+		if (!pwd) break;
+		if (strnpathcmp(name, pwd -> pw_name, len)) continue;
+		size = strlen(pwd -> pw_name);
+		new = malloc2(size + 2 + 1);
+		new[0] = '~';
+		strcatdelim2(new + 1, pwd -> pw_name, NULL);
+		if (finddupl(new, argc, *argvp)) {
+			free(new);
+			continue;
+		}
+
+		*argvp = (char **)realloc2(*argvp,
+			(argc + 1) * sizeof(char *));
+		(*argvp)[argc++] = new;
+	}
+	_mtrace_file = "endpwent(start)";
+	endpwent();
+	if (_mtrace_file) _mtrace_file = NULL;
+	else {
+		_mtrace_file = "endpwent(end)";
+		malloc(0);	/* dummy alloc */
+	}
+# else	/* !DEBUG */
+	setpwent();
+	while ((pwd = getpwent())) {
 		if (strnpathcmp(name, pwd -> pw_name, len)) continue;
 		size = strlen(pwd -> pw_name);
 		new = malloc2(size + 2 + 1);
@@ -1702,22 +1841,22 @@ char ***argvp;
 		(*argvp)[argc++] = new;
 	}
 	endpwent();
+# endif	/* !DEBUG */
 	return(argc);
 }
 #endif	/* !MSDOS */
 
-static int NEAR completefile(file, dir, dlen, exe, argc, argvp)
-char *file, *dir;
-int dlen, exe, argc;
-char ***argvp;
+static int NEAR completefile(file, len, argc, argvp, dir, dlen, exe)
+char *file;
+int len, argc;
+char ***argvp, *dir;
+int dlen, exe;
 {
 	DIR *dirp;
 	struct dirent *dp;
-	struct stat st;
 	char *cp, *new, path[MAXPATHLEN];
-	int len, size;
+	int d, size;
 
-	len = strlen(file);
 	if (dlen >= MAXPATHLEN - 2) return(argc);
 	strncpy2(path, dir, dlen);
 	if (!(dirp = Xopendir(path))) return(argc);
@@ -1730,13 +1869,11 @@ char ***argvp;
 		if (size + (cp - path) >= MAXPATHLEN) continue;
 		strncpy2(cp, dp -> d_name, size);
 
-		if (Xstat(path, &st) < 0
-		|| (exe && (st.st_mode & S_IFMT) != S_IFDIR
-		&& Xaccess(path, X_OK) < 0)) continue;
+		if ((d = isexecute(path, 1, exe)) < 0) continue;
 
 		new = malloc2(size + 1 + 1);
 		strncpy(new, dp -> d_name, size);
-		if ((st.st_mode & S_IFMT) == S_IFDIR) new[size++] = _SC_;
+		if (d) new[size++] = _SC_;
 		new[size] = '\0';
 		if (finddupl(new, argc, *argvp)) {
 			free(new);
@@ -1751,57 +1888,63 @@ char ***argvp;
 	return(argc);
 }
 
-static int NEAR completeexe(file, argc, argvp)
+static int NEAR completeexe(file, len, argc, argvp)
 char *file;
-int argc;
+int len, argc;
 char ***argvp;
 {
-	char *cp, *path;
-	int len;
+	char *cp, *tmp, *next;
+	int dlen;
 
 #if	MSDOS && !defined (DISMISS_CURPATH)
-	argc = completefile(file, ".", 1, 1, argc, argvp);
+	argc = completefile(file, len, argc, argvp, ".", 1, 1);
 #endif
-	if (!(path = getvar("PATH", -1))) return(argc);
-	do {
+	if (!(next = getvar("PATH", -1))) return(argc);
+	for (cp = next; cp; cp = next) {
 #if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
-		if (_dospath(path)) cp = strchr(path + 2, PATHDELIM);
+		if (_dospath(cp)) next = strchr(cp + 2, PATHDELIM);
 		else
 #endif
-		cp = strchr(path, PATHDELIM);
-		len = (cp) ? cp++ - path : strlen(path);
-		argc = completefile(file, path, len, 1, argc, argvp);
-		path = cp;
-	} while(path);
+		next = strchr(cp, PATHDELIM);
+		dlen = (next) ? (next++) - cp : strlen(cp);
+		tmp = _evalpath(cp, cp + dlen, 1, 1);
+		dlen = strlen(tmp);
+		argc = completefile(file, len, argc, argvp, tmp, dlen, 1);
+		free(tmp);
+	}
 	return(argc);
 }
 
-int completepath(path, exe, argc, argvp)
+int completepath(path, len, argc, argvp, exe)
 char *path;
-int exe, argc;
+int len, argc;
 char ***argvp;
+int exe;
 {
 	char *file, *dir;
-	int len;
+	int dlen;
 
 	dir = path;
-	len = 0;
+	dlen = 0;
 #if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
 	if (_dospath(path)) {
 		dir += 2;
-		len = 2;
+		dlen = 2;
 	}
 #endif
 
 	if ((file = strrdelim(dir, 0))) {
-		len += (file == dir) ? 1 : file - dir;
-		return(completefile(file + 1, path, len, exe, argc, argvp));
+		dlen += (file == dir) ? 1 : file - dir;
+		return(completefile(file + 1, strlen(file + 1), argc, argvp,
+			path, dlen, exe));
 	}
 #if	!MSDOS
-	else if (*path == '~') return(completeuser(path + 1, argc, argvp));
+	else if (*path == '~')
+		return(completeuser(path + 1, len - 1, argc, argvp));
 #endif
-	else if (exe && dir == path) return(completeexe(path, argc, argvp));
-	return(completefile(path, ".", 1, exe, argc, argvp));
+	else if (exe && dir == path)
+		return(completeexe(path, len, argc, argvp));
+	return(completefile(path, len, argc, argvp, ".", 1, exe));
 }
 
 char *findcommon(argc, argv)
@@ -1825,9 +1968,9 @@ char **argv;
 	}
 	return(common);
 }
-#endif	/* !FDSH && !_NOCOMPLETE */
+#endif	/* !_NOCOMPLETE */
 
-int addmeta(s1, s2, stripm, quoted)
+static int NEAR addmeta(s1, s2, stripm, quoted)
 char *s1, *s2;
 int stripm, quoted;
 {
@@ -1865,6 +2008,26 @@ int stripm, quoted;
 	return(j);
 }
 
+char *catvar(argv, delim)
+char *argv[];
+int delim;
+{
+	char *cp;
+	int i, len;
+
+	if (!argv) return(NULL);
+	for (i = len = 0; argv[i]; i++) len += strlen(argv[i]);
+	if (i < 1) return(strdup2(""));
+	len += (delim) ? i - 1 : 0;
+	cp = malloc2(len + 1);
+	len = strcpy2(cp, argv[0]) - cp;
+	for (i = 1; argv[i]; i++) {
+		if (delim) cp[len++] = delim;
+		len = strcpy2(cp + len, argv[i]) - cp;
+	}
+	return(cp);
+}
+
 static char *NEAR evalshellparam(c, quoted)
 int c, quoted;
 {
@@ -1879,10 +2042,11 @@ int c, quoted;
 			|| !(arglist = (*getarglistfunc)())) break;
 			sp = ' ';
 			if (!quoted) {
-				cp = catargs(&(arglist[1]), sp);
+				cp = catvar(&(arglist[1]), sp);
 				break;
 			}
 #ifdef	BASHSTYLE
+	/* bash uses IFS instead of a space as a separator */
 			if ((cp = getvar("IFS", -1)) && *cp) sp = *cp;
 #endif
 
@@ -1909,10 +2073,11 @@ int c, quoted;
 			|| !(arglist = (*getarglistfunc)())) break;
 			sp = ' ';
 #ifdef	BASHSTYLE
+	/* bash uses IFS instead of a space as a separator */
 			if (quoted && (cp = getvar("IFS", -1)))
 				sp = (*cp) ? *cp : '\0';
 #endif
-			cp = catargs(&(arglist[1]), sp);
+			cp = catvar(&(arglist[1]), sp);
 			break;
 		case '#':
 			if (getarglistfunc
@@ -1965,7 +2130,7 @@ int s, len, vlen, mode;
 
 	val = malloc2(vlen + 1);
 	strncpy2(val, arg + s, vlen);
-	*cpp = evalarg(val, (mode == '=' || mode == '?'));
+	*cpp = evalarg(val, (mode == '=' || mode == '?'), 0);
 	free(val);
 	if (!(*cpp)) return(-1);
 
@@ -1976,10 +2141,11 @@ int s, len, vlen, mode;
 			return(-1);
 		}
 #ifdef	BASHSTYLE
+	/* bash does not evaluates a quoted string in substitution itself */
 		free(*cpp);
 		val = malloc2(vlen + 1);
 		strncpy2(val, arg + s, vlen);
-		*cpp = evalarg(val, 0);
+		*cpp = evalarg(val, 0, 0);
 		free(val);
 		if (!(*cpp)) return(-1);
 #endif
@@ -2042,7 +2208,7 @@ int quoted;
 	}
 	else {
 		nest = 1;
-		len = s = e = -1;
+		len = s = -1;
 		top++;
 		quote = '\0';
 		for (i = 0; nest > 0; i++) {
@@ -2059,9 +2225,6 @@ int quoted;
 			else if (quote == '\'');
 			else if (ismeta(top, i, quote)) {
 				if (len < 0) return(-1);
-#ifndef	BASHSTYLE
-				if (!quote) e = i;
-#endif
 			}
 			else if (quote);
 			else if (top[i] == '\'' || top[i] == '"') {
@@ -2097,20 +2260,14 @@ int quoted;
 					mode = top[i];
 					break;
 				default:
-					if (len >= 0) {
-#ifndef	BASHSTYLE
-						if (!quoted
-						&& strchr(IFS_SET, top[i]))
-							return(-1);
-#endif
-					}
+					if (len >= 0);
 					else if (i > 0 && top[i] != '_'
 					&& !isalnum(top[i]))
 						return(-1);
 					break;
 			}
 		}
-		if (e < 0) e = i - 1;
+		e = i - 1;
 		if (len < 0) len = e;
 		if (s < 0) s = e;
 		*argp += i + 1;
@@ -2175,16 +2332,133 @@ int quoted;
 	return(ptr);
 }
 
+#if	!MSDOS
+uidtable *finduid(uid, name)
+uid_t uid;
+char *name;
+{
+	struct passwd *pwd;
+	int i;
+
+	if (name) {
+		for (i = 0; i < maxuid; i++)
+			if (!strpathcmp(name, uidlist[i].name))
+				return(&(uidlist[i]));
+# ifdef	DEBUG
+		_mtrace_file = "getpwnam(start)";
+		pwd = getpwnam(name);
+		if (_mtrace_file) _mtrace_file = NULL;
+		else {
+			_mtrace_file = "getpwnam(end)";
+			malloc(0);	/* dummy alloc */
+		}
+# else
+		pwd = getpwnam(name);
+# endif
+	}
+	else {
+		for (i = 0; i < maxuid; i++)
+			if (uid == uidlist[i].uid) return(&(uidlist[i]));
+# ifdef	DEBUG
+		_mtrace_file = "getpwuid(start)";
+		pwd = getpwuid(uid);
+		if (_mtrace_file) _mtrace_file = NULL;
+		else {
+			_mtrace_file = "getpwuid(end)";
+			malloc(0);	/* dummy alloc */
+		}
+# else
+		pwd = getpwuid(uid);
+# endif
+	}
+
+	if (!pwd) return(NULL);
+	uidlist = b_realloc(uidlist, maxuid, uidtable);
+	uidlist[maxuid].uid = pwd -> pw_uid;
+	uidlist[maxuid].name = strdup2(pwd -> pw_name);
+	uidlist[maxuid].home = strdup2(pwd -> pw_dir);
+	return(&(uidlist[maxuid++]));
+}
+
+gidtable *findgid(gid, name)
+gid_t gid;
+char *name;
+{
+	struct group *grp;
+	int i;
+
+	if (name) {
+		for (i = 0; i < maxgid; i++)
+			if (!strpathcmp(name, gidlist[i].name))
+				return(&(gidlist[i]));
+# ifdef	DEBUG
+		_mtrace_file = "getgrnam(start)";
+		grp = getgrnam(name);
+		if (_mtrace_file) _mtrace_file = NULL;
+		else {
+			_mtrace_file = "getgrnam(end)";
+			malloc(0);	/* dummy alloc */
+		}
+# else
+		grp = getgrnam(name);
+# endif
+	}
+	else {
+		for (i = 0; i < maxgid; i++)
+			if (gid == gidlist[i].gid) return(&(gidlist[i]));
+# ifdef	DEBUG
+		_mtrace_file = "getgrgid(start)";
+		grp = getgrgid(gid);
+		if (_mtrace_file) _mtrace_file = NULL;
+		else {
+			_mtrace_file = "getgrgid(end)";
+			malloc(0);	/* dummy alloc */
+		}
+# else
+		grp = getgrgid(gid);
+# endif
+	}
+
+	if (!grp) return(NULL);
+	gidlist = b_realloc(gidlist, maxgid, gidtable);
+	gidlist[maxgid].gid = grp -> gr_gid;
+	gidlist[maxgid].name = strdup2(grp -> gr_name);
+	return(&(gidlist[maxgid++]));
+}
+
+# ifdef	DEBUG
+VOID freeidlist(VOID_A)
+{
+	int i;
+
+	if (uidlist) {
+		for (i = 0; i < maxuid; i++) {
+			free(uidlist[i].name);
+			free(uidlist[i].home);
+		}
+		free(uidlist);
+	}
+	if (gidlist) {
+		for (i = 0; i < maxgid; i++) free(gidlist[i].name);
+		free(gidlist);
+	}
+	uidlist = NULL;
+	gidlist = NULL;
+	maxuid = maxgid = 0;
+}
+# endif
+#endif	/* !MSDOS */
+
 char *gethomedir(VOID_A)
 {
 #if	!MSDOS
-	struct passwd *pwd;
+	uidtable *up;
 #endif
 	char *cp;
 
 	if ((cp = getvar("HOME", -1))) return(cp);
 #if	!MSDOS
-	if ((pwd = getpwuid(getuid()))) return(pwd -> pw_dir);
+	if ((up = finduid(getuid(), NULL))) return(up -> home);
 #endif
 	return(NULL);
 }
@@ -2207,13 +2481,13 @@ char **argp;
 #endif
 	else {
 #if	!MSDOS
-		struct passwd *pwd;
+		uidtable *up;
 
 		cp = malloc2(len + 1);
 		strncpy2(cp, top, len);
-		pwd = getpwnam(cp);
+		up = finduid(0, cp);
 		free(cp);
-		if (pwd) cp = pwd -> pw_dir;
+		if (up) cp = up -> home;
 		else
 #endif
 		cp = NULL;
@@ -2234,56 +2508,112 @@ char **argp;
 	return(ptr);
 }
 
-char *evalarg(arg, stripq)
+char *evalarg(arg, stripq, backq)
 char *arg;
-int stripq;
+int stripq, backq;
 {
-	char *cp, *buf;
-	int i, prev, quote;
+#if	defined (FD) && MSDOS && !defined (_NOUSELFN)
+	char path[MAXPATHLEN], alias[MAXPATHLEN];
+#endif
+	char *cp, *buf, *bbuf;
+	int i, j, prev, quote;
 
-	buf = malloc2(strlen(arg) + 1);
+#if	defined (FD) && MSDOS && !defined (_NOUSELFN)
+	if (*arg == '"' && (cp = strchr(arg + 1, '"')) && !*(cp + 1)) {
+		strncpy2(path, arg + 1, cp - arg - 1);
+		if (shortname(path, alias) && strpathcmp(path, alias)) {
+			if (stripq) return(strdup2(alias));
+			i = strlen(alias);
+			buf = malloc2(i + 2 + 1);
+			buf[0] = '"';
+			memcpy(&(buf[1]), alias, i);
+			buf[++i] = '"';
+			buf[++i] = '\0';
+			return(buf);
+		}
+	}
+#endif
+	i = strlen(arg) + 1;
+	buf = malloc2(i);
+	bbuf = (backq) ? malloc2(i) : NULL;
 	prev = quote = '\0';
 	i = 0;
+	j = -1;
 	for (cp = arg; *cp; prev = *(cp++)) {
 		if (*cp == quote) {
 			quote = '\0';
-			if (!stripq) buf[i++] = *cp;
+			if (j >= 0) bbuf[j++] = *cp;
+			else if (!stripq) buf[i++] = *cp;
 			continue;
 		}
 #ifdef	CODEEUC
-		else if (isekana(cp, 0)) buf[i++] = *(cp++);
+		else if (isekana(cp, 0)) {
+			if (j >= 0) bbuf[j++] = *cp;
+			else buf[i++] = *cp;
+			cp++;
+		}
 #endif
-		else if (iskanji1(cp, 0)) buf[i++] = *(cp++);
+		else if (iskanji1(cp, 0)) {
+			if (j >= 0) bbuf[j++] = *cp;
+			buf[i++] = *cp;
+			cp++;
+		}
 		else if (quote == '\'');
-		else if (*cp == '$' && *(cp + 1)) {
+		else if (*cp == '$' && *(cp + 1) && j < 0) {
 #ifdef	FAKEMETA
-			if (*(cp + 1) == '$' || *(cp + 1) == '~') {
-				buf[i++] = *(++cp);
-				continue;
-			}
+			if (*(cp + 1) == '$' || *(cp + 1) == '~') cp++;
+			else
 #endif
-			if ((i = evalvar(&buf, i, &cp, quote)) < 0) {
+			if ((i = evalvar(&buf, i, &cp, quote)) >= 0) continue;
+			else {
+				if (bbuf) free(bbuf);
 				free(buf);
 				if (i < -1) *arg = '\0';
 				return(NULL);
 			}
+		}
+		else if (ismeta(cp, 0, quote)) {
+			if (j >= 0);
+			else buf[i++] = *cp;
+			cp++;
+		}
+		else if (backq && backquotefunc && *cp == '`') {
+			if (j < 0) j = 0;
+			else if (j) {
+				char *tmp;
+				int size;
+
+				bbuf[j] = '\0';
+				tmp = (*backquotefunc)(bbuf);
+				if (tmp) {
+					j = addmeta(NULL, tmp, 1, 0);
+					size = i + j + strlen(cp + 1) + 1;
+					buf = realloc2(buf, size);
+					addmeta(&(buf[i]), tmp, 1, 0);
+					i += j;
+					free(tmp);
+				}
+				j = -1;
+			}
 			continue;
 		}
-		else if (ismeta(cp, 0, quote)) buf[i++] = *(cp++);
 		else if (quote);
 		else if (*cp == '\'' || *cp == '"') {
 			quote = *cp;
-			if (!stripq) buf[i++] = *cp;
+			if (j >= 0) bbuf[j++] = *cp;
+			else if (!stripq) buf[i++] = *cp;
 			continue;
 		}
-		else if (*cp == '~'
+		else if (*cp == '~' && j < 0
 		&& (!prev || prev == ':' || prev == '=')) {
 			i = evalhome(&buf, i, &cp);
 			continue;
 		}
 
-		buf[i++] = *cp;
+		if (j >= 0) bbuf[j++] = *cp;
+		else buf[i++] = *cp;
 	}
+	if (bbuf) free(bbuf);
 	buf[i] = '\0';
 	return(buf);
 }
@@ -2310,9 +2640,20 @@ int iscom;
 			|| (*argvp)[n][i] == '"')
 				quote = (*argvp)[n][i];
 			else if (strchr(ifs, (*argvp)[n][i])) {
-				(*argvp)[n][i++] = '\0';
-				if (!(*argvp)[n][i]) break;
-				cp = strdup2(&((*argvp)[n][i]));
+				for (j = i + 1; (*argvp)[n][j]; j++)
+					if (!strchr(ifs, (*argvp)[n][j]))
+						break;
+				if (!i) {
+					for (i = 0; (*argvp)[n][i + j]; i++)
+						(*argvp)[n][i] =
+							(*argvp)[n][i + j];
+					(*argvp)[n][i] = '\0';
+					i = -1;
+					continue;
+				}
+				(*argvp)[n][i] = '\0';
+				if (!(*argvp)[n][j]) break;
+				cp = strdup2(&((*argvp)[n][j]));
 				*argvp = (char **)realloc2(*argvp,
 					(argc + 2) * sizeof(char *));
 				for (j = argc; j > n; j--)
@@ -2333,10 +2674,10 @@ int iscom;
 	return(argc);
 }
 
-int evalglob(argc, argvp, iscom)
+int evalglob(argc, argvp, stripq, iscom)
 int argc;
 char ***argvp;
-int iscom;
+int stripq, iscom;
 {
 	char *cp, **wild;
 	int i, j, n, w, size, quote;
@@ -2347,7 +2688,7 @@ int iscom;
 			cp = c_realloc(cp, j + 1, size);
 			if ((*argvp)[n][i] == quote) {
 				quote = '\0';
-				continue;
+				if (stripq) continue;
 			}
 #ifdef	CODEEUC
 			else if (isekana((*argvp)[n], i))
@@ -2355,25 +2696,18 @@ int iscom;
 #endif
 			else if (iskanji1((*argvp)[n], i))
 				cp[j++] = (*argvp)[n][i++];
-			else if (quote == '\'') {
-#ifdef	EVALMACRO
-				if ((*argvp)[n][i] == '%') {
-					cp = c_realloc(cp, j + 2, size);
-					cp[j++] = (*argvp)[n][i];
-				}
-#endif
-			}
+			else if (quote == '\'');
 			else if (ismeta((*argvp)[n], i, quote)) {
 				i++;
-				if (quote && (*argvp)[n][i] != quote
-				&& (*argvp)[n][i] != PMETA)
+				if (quote
+				&& !strchr(DQ_METACHAR, (*argvp)[n][i]))
 					cp[j++] = PMETA;
 			}
 			else if (quote);
 			else if ((*argvp)[n][i] == '\''
 			|| (*argvp)[n][i] == '"') {
 				quote = (*argvp)[n][i];
-				continue;
+				if (stripq) continue;
 			}
 			else if (iscom && !n);
 			else if (!strchr("?*[", (*argvp)[n][i]));
@@ -2415,50 +2749,43 @@ int iscom;
 	return(argc);
 }
 
-char *stripquote(arg)
+int stripquote(arg, stripq)
 char *arg;
+int stripq;
 {
-	int i, j, quote;
+	int i, j, quote, stripped;
 
+	stripped = 0;
 	for (i = j = 0, quote = '\0'; arg[i]; i++) {
 		if (arg[i] == quote) {
 			quote = '\0';
-			continue;
+			stripped = 1;
+			if (stripq) continue;
 		}
 #ifdef	CODEEUC
 		else if (isekana(arg, i)) arg[j++] = arg[i++];
 #endif
 		else if (iskanji1(arg, i)) arg[j++] = arg[i++];
-		else if (quote == '\'') {
-#ifdef	EVALMACRO
-			if (arg[i] == '%') {
-				int k;
-
-				k = strlen(&(arg[i])) + 1;
-				arg = realloc2(arg, i + k + 1);
-				while (--k > 0) arg[j + k + 1] = arg[i + k];
-				arg[j++] = arg[i];
-			}
-#endif
-		}
+		else if (quote == '\'');
 		else if (ismeta(arg, i, quote)) {
 			i++;
+			stripped = 1;
 			if (quote && arg[i] != quote && arg[i] != PMETA)
 				arg[j++] = PMETA;
 		}
 		else if (quote);
 		else if (arg[i] == '\'' || arg[i] == '"') {
 			quote = arg[i];
-			continue;
+			stripped = 1;
+			if (stripq) continue;
 		}
 
 		arg[j++] = arg[i];
 	}
 	arg[j] = '\0';
-	return(arg);
+	return(stripped);
 }
 
-#ifndef	FDSH
 char *_evalpath(path, eol, uniqdelim, evalq)
 char *path, *eol;
 int uniqdelim, evalq;
@@ -2475,7 +2802,7 @@ int uniqdelim, evalq;
 	cp = malloc2(i + 1);
 	strncpy2(cp, path, i);
 
-	if (!(tmp = evalarg(cp, 0))) {
+	if (!(tmp = evalarg(cp, 0, 0))) {
 		*cp = '\0';
 		return(cp);
 	}
@@ -2511,7 +2838,12 @@ int uniqdelim, evalq;
 #endif
 		else if (iskanji1(cp, i)) tmp[j++] = cp[i++];
 		else if (quote == '\'');
-		else if (ismeta(cp, i, quote)) tmp[j++] = cp[i++];
+		else if (ismeta(cp, i, quote)) {
+			i++;
+			if (!evalq
+			|| (quote && !strchr(DQ_METACHAR, cp[i])))
+				tmp[j++] = PMETA;
+		}
 		else if (quote);
 		else if (cp[i] == '\'' || cp[i] == '"') {
 #if	defined (FD) && MSDOS && !defined (_NOUSELFN)
@@ -2540,4 +2872,3 @@ int uniqdelim;
 	free(path);
 	return(cp);
 }
-#endif	/* !FDSH */
