@@ -4,6 +4,31 @@
  *	Variables for TERMCAP
  */
 
+#ifndef	__TERM_H_
+#define	__TERM_H_
+
+#ifndef	VOID_P
+#ifdef	NOVOID
+#define	VOID
+#define	VOID_T	int
+#define	VOID_P	char *
+#else
+#define	VOID	void
+#define	VOID_T	void
+#define	VOID_P	void *
+#endif
+#endif
+
+#ifndef	STDIN_FILENO
+#define	STDIN_FILENO	0
+#endif
+#ifndef	STDOUT_FILENO
+#define	STDOUT_FILENO	1
+#endif
+#ifndef	STDERR_FILENO
+#define	STDERR_FILENO	2
+#endif
+
 #define	CR	'\r'
 #define	ESC	'\033'
 
@@ -66,6 +91,9 @@ extern int n_line;
 extern int stable_standout;
 extern char *t_init;
 extern char *t_end;
+extern char *t_metamode;
+extern char *t_nometamode;
+extern char *t_scroll;
 extern char *t_keypad;
 extern char *t_nokeypad;
 extern char *t_normalcursor;
@@ -96,23 +124,36 @@ extern char *c_home;
 extern char *c_locate;
 extern char *c_return;
 extern char *c_newline;
+extern char *c_scrollforw;
+extern char *c_scrollrev;
 extern char *c_up;
 extern char *c_down;
 extern char *c_right;
 extern char *c_left;
+extern char *c_nup;
+extern char *c_ndown;
+extern char *c_nright;
+extern char *c_nleft;
 extern u_char cc_intr;
 extern u_char cc_quit;
 extern u_char cc_eof;
 extern u_char cc_eol;
+extern VOID_T (*keywaitfunc)__P_((VOID_A));
+#if	!MSDOS
+extern int usegetcursor;
+#endif
+extern int ttyio;
+extern FILE *ttyout;
 
 #if	MSDOS
-#define	putterm(str)	cputs2(str)
-#define	putterms(str)	cputs2(str)
+#define	putterm(s)	cputs2(s)
+#define	putterms(s)	cputs2(s)
 #else
-#define	putterm(str)	tputs(str, 1, putch2)
-#define	putterms(str)	tputs(str, n_line, putch2)
+#define	putterm(s)	tputs(s, 1, putch3)
+#define	putterms(s)	tputs(s, n_line, putch3)
 #endif
 
+extern int opentty __P_((VOID_A));
 extern int inittty __P_((int));
 extern int cooked2 __P_((VOID_A));
 extern int cbreak2 __P_((VOID_A));
@@ -124,25 +165,29 @@ extern int nonl2 __P_((VOID_A));
 extern int tabs __P_((VOID_A));
 extern int notabs __P_((VOID_A));
 extern int keyflush __P_((VOID_A));
+extern int ttyiomode __P_((VOID_A));
+extern int stdiomode __P_((VOID_A));
 extern int exit2 __P_((int));
+extern int getxy __P_((int *, int *));
+extern char *tparamstr __P_((char *, int, int));
 extern int getterment __P_((VOID_A));
 #if	!MSDOS
-extern int setkeyseq __P_((int, char *));
-extern char *getkeyseq __P_((int));
+extern int setkeyseq __P_((int, char *, int));
+extern char *getkeyseq __P_((int, int *));
 #endif
 extern int initterm __P_((VOID_A));
 extern int endterm __P_((VOID_A));
 extern int putch2 __P_((int));
-extern int cputs2 __P_((char *));
-#if	MSDOS || defined (__STDC__)
-extern int cprintf2(CONST char *, ...);
-#else
-extern int cprintf2 __P_((CONST char *, ...));
+#if	!MSDOS
+extern int putch3 __P_((int));
 #endif
+extern int cputs2 __P_((char *));
+extern int cprintf2 __P_((CONST char *, ...));
 extern int kbhit2 __P_((u_long));
 extern int getch2 __P_((VOID_A));
 extern int getkey2 __P_((int));
 extern int ungetch2 __P_((u_char));
+extern int setscroll __P_((int, int));
 extern int locate __P_((int, int));
 extern int tflush __P_((VOID_A));
 extern int getwsize __P_((int, int));
@@ -154,6 +199,9 @@ extern int chgcolor __P_((int, int));
 #ifndef	WAITKEYPAD
 #define	WAITKEYPAD	360		/* msec */
 #endif
+#ifndef	WAITMETA
+#define	WAITMETA	60		/* msec */
+#endif
 
 #define	ANSI_BLACK	0
 #define	ANSI_RED	1
@@ -165,3 +213,5 @@ extern int chgcolor __P_((int, int));
 #define	ANSI_WHITE	7
 #define	ANSI_NORMAL	30
 #define	ANSI_REVERSE	40
+
+#endif	/* __TERM_H_ */
