@@ -23,7 +23,8 @@ extern int preparedrv __P_((int));
 
 
 extern char fullpath[];
-extern int sorton;
+extern int win_x;
+extern int win_y;
 #ifndef	_NODOSDRIVE
 extern int lastdrv;
 #endif
@@ -47,14 +48,13 @@ static int NEAR expandtree __P_((treelist *));
 static int NEAR expandall __P_((treelist *));
 static int NEAR treeup __P_((VOID_A));
 static int NEAR treedown __P_((VOID_A));
-static VOID NEAR freetree __P_((treelist *, int));
+static int NEAR freetree __P_((treelist *, int));
 static VOID NEAR _tree_search __P_((VOID_A));
 static int NEAR _tree_input __P_((VOID_A));
 static char *NEAR _tree __P_((VOID_A));
 
 int sorttree = 0;
 int dircountlimit = 0;
-char *treepath = NULL;
 
 static int redraw = 0;
 static int tr_no = 0;
@@ -329,7 +329,6 @@ static VOID NEAR treebar(VOID_A)
 	kanjiputs2(treepath, n_column - 6, 0);
 	locate(0, LMESLINE);
 	putterm(l_clear);
-	locate(0, 0);
 }
 
 VOID rewritetree(VOID_A)
@@ -507,17 +506,19 @@ static int NEAR treedown(VOID_A)
 	return(0);
 }
 
-static VOID NEAR freetree(list, max)
+static int NEAR freetree(list, max)
 treelist *list;
 int max;
 {
-	int i;
+	int i, n;
 
-	for (i = 0; i < max; i++) {
+	for (i = n = 0; i < max; i++) {
+		n++;
 		if (list[i].name) free(list[i].name);
-		if (list[i].sub) freetree(list[i].sub, list[i].max);
+		if (list[i].sub) n += freetree(list[i].sub, list[i].max);
 	}
 	free(list);
+	return(n);
 }
 
 static VOID NEAR _tree_search(VOID_A)
@@ -631,13 +632,15 @@ static int NEAR _tree_input(VOID_A)
 		case K_LEFT:
 			if (tr_cur && (tr_cur -> sub)
 			&& (tr_cur -> sub[tr_no]).sub) {
-				freetree((tr_cur -> sub[tr_no]).sub,
+				tmp = freetree((tr_cur -> sub[tr_no]).sub,
 					(tr_cur -> sub[tr_no]).max);
+				tr_bottom -= tmp;
 				(tr_cur -> sub[tr_no]).max = -1;
 				(tr_cur -> sub[tr_no]).sub = NULL;
 				redraw = 1;
 				break;
 			}
+/*FALLTHRU*/
 		case K_BS:
 			do {
 				if (treeup() < 0) break;
@@ -657,7 +660,7 @@ static int NEAR _tree_input(VOID_A)
 		case CTRL('L'):
 			rewritefile(1);
 			break;
-		case ESC:
+		case K_ESC:
 			break;
 		default:
 			if (ch < 'A' || ch > 'Z') break;
@@ -742,11 +745,14 @@ static char *NEAR _tree(VOID_A)
 	tr_scr = malloc2((FILEPERLOW - 1) * (TREEFIELD + 1));
 	searchtree();
 	showtree();
+	win_x = 0;
 	do {
 		treebar();
 		oy = tr_line;
 		otop = tr_top;
 		redraw = 0;
+		win_y = tr_line;
+		locate(win_x, win_y);
 		tflush();
 
 		if ((ch = _tree_input()) == 'l') {
@@ -766,12 +772,12 @@ static char *NEAR _tree(VOID_A)
 			else kanjiputs2(bufptr(oy), TREEFIELD, 0);
 			evaldir(path, 1);
 		}
-	} while (ch != ESC && ch != CR);
+	} while (ch != K_ESC && ch != K_CR);
 
 	free(tr_scr);
 	freetree(tr_root, 1);
 	treepath = NULL;
-	if (ch == ESC) return(NULL);
+	if (ch == K_ESC) return(NULL);
 	return(strdup2(path));
 }
 
